@@ -36,8 +36,11 @@
 #include "got_error.h"
 #include "got_path.h"
 #include "got_object.h"
+#include "got_reference.h"
 
 #include "gotsysd.h"
+#include "media.h"
+#include "gotwebd.h"
 #include "gotsys.h"
 #include "log.h"
 #include "helpers.h"
@@ -174,8 +177,8 @@ get_helper_prog_name(int imsg_type)
 		return GOTSYSD_PATH_PROG_WRITE_CONF;
 	case GOTSYSD_IMSG_START_PROG_APPLY_CONF:
 		return GOTSYSD_PATH_PROG_APPLY_CONF;
-	case GOTSYSD_IMSG_START_PROG_SSHDCONFIG:
-		return GOTSYSD_PATH_PROG_SSHDCONFIG;
+	case GOTSYSD_IMSG_START_PROG_APPLY_WEBCONF:
+		return GOTSYSD_PATH_PROG_APPLY_WEBCONF;
 	default:
 		return NULL;
 	}
@@ -441,12 +444,12 @@ send_apply_conf_ready(struct gotsysd_imsgev *iev)
 }
 
 static const struct got_error *
-send_sshdconfig_ready(struct gotsysd_imsgev *iev)
+send_apply_webconf_ready(struct gotsysd_imsgev *iev)
 {
 	if (gotsysd_imsg_compose_event(iev,
-	    GOTSYSD_IMSG_SYSCONF_SSHDCONFIG_READY, gotsysd_helpers.proc_id,
+	    GOTSYSD_IMSG_SYSCONF_APPLY_WEBCONF_READY, gotsysd_helpers.proc_id,
 	    -1, NULL, 0) == -1)
-		return got_error_from_errno("imsg_compose SSHDCONFIG_READY");
+		return got_error_from_errno("imsg_compose APPLY_WEBCONF_READY");
 	
 	return NULL;
 }
@@ -490,8 +493,8 @@ proc_ready(struct gotsysd_helper_proc *proc)
 	case GOTSYSD_IMSG_START_PROG_APPLY_CONF:
 		err = send_apply_conf_ready(sysconf_iev);
 		break;
-	case GOTSYSD_IMSG_START_PROG_SSHDCONFIG:
-		err = send_sshdconfig_ready(sysconf_iev);
+	case GOTSYSD_IMSG_START_PROG_APPLY_WEBCONF:
+		err = send_apply_webconf_ready(sysconf_iev);
 		break;
 	default:
 		err = got_error_fmt(GOT_ERR_PRIVSEP_MSG,
@@ -577,10 +580,27 @@ dispatch_helper_child(int fd, short event, void *arg)
 		case GOTSYSD_IMSG_SYSCONF_GROUP_MEMBERS:
 		case GOTSYSD_IMSG_SYSCONF_GROUP_MEMBERS_DONE:
 		case GOTSYSD_IMSG_SYSCONF_GROUPS_DONE:
+		case GOTSYSD_IMSG_SYSCONF_WEB_SERVER:
+		case GOTSYSD_IMSG_SYSCONF_WEB_SERVERS_DONE:
+		case GOTSYSD_IMSG_SYSCONF_WEB_ACCESS_RULE:
+		case GOTSYSD_IMSG_SYSCONF_WEB_ACCESS_RULES_DONE:
 		case GOTSYSD_IMSG_SYSCONF_REPO:
+		case GOTSYSD_IMSG_SYSCONF_WEBREPO:
+		case GOTSYSD_IMSG_SYSCONF_WEBREPO_ACCESS_RULE:
+		case GOTSYSD_IMSG_SYSCONF_WEBREPO_ACCESS_RULES_DONE:
+		case GOTSYSD_IMSG_SYSCONF_WEBREPOS_DONE:
+		case GOTSYSD_IMSG_SYSCONF_WEBSITE_PATH:
+		case GOTSYSD_IMSG_SYSCONF_WEBSITE:
+		case GOTSYSD_IMSG_SYSCONF_WEBSITE_ACCESS_RULE:
+		case GOTSYSD_IMSG_SYSCONF_WEBSITE_ACCESS_RULES_DONE:
+		case GOTSYSD_IMSG_SYSCONF_WEBSITES_DONE:
+		case GOTSYSD_IMSG_SYSCONF_GLOBAL_MEDIA_TYPE:
+		case GOTSYSD_IMSG_SYSCONF_GLOBAL_MEDIA_TYPES_DONE:
 		case GOTSYSD_IMSG_SYSCONF_REPOS_DONE:
 		case GOTSYSD_IMSG_SYSCONF_ACCESS_RULE:
 		case GOTSYSD_IMSG_SYSCONF_ACCESS_RULES_DONE:
+		case GOTSYSD_IMSG_SYSCONF_MEDIA_TYPE:
+		case GOTSYSD_IMSG_SYSCONF_MEDIA_TYPES_DONE:
 		case GOTSYSD_IMSG_SYSCONF_PROTECTED_TAG_NAMESPACES:
 		case GOTSYSD_IMSG_SYSCONF_PROTECTED_TAG_NAMESPACES_ELEM:
 		case GOTSYSD_IMSG_SYSCONF_PROTECTED_BRANCH_NAMESPACES:
@@ -711,9 +731,9 @@ dispatch_helper_child(int fd, short event, void *arg)
 				err = got_error_from_errno("imsg_forward");
 			shut = 1;
 			break;
-		case GOTSYSD_IMSG_SYSCONF_INSTALL_SSHD_CONFIG_DONE:
+		case GOTSYSD_IMSG_SYSCONF_APPLY_WEBCONF_DONE:
 			if (proc->type !=
-			    GOTSYSD_IMSG_START_PROG_SSHDCONFIG) {
+			    GOTSYSD_IMSG_START_PROG_APPLY_WEBCONF) {
 				err = got_error_fmt(GOT_ERR_PRIVSEP_MSG,
 				    "unexpected message type %d from helper "
 				        "process type %d pid %u\n",
@@ -911,7 +931,7 @@ helpers_dispatch_sysconf(int fd, short event, void *arg)
 		case GOTSYSD_IMSG_START_PROG_GROUPADD:
 		case GOTSYSD_IMSG_START_PROG_WRITE_CONF:
 		case GOTSYSD_IMSG_START_PROG_APPLY_CONF:
-		case GOTSYSD_IMSG_START_PROG_SSHDCONFIG:
+		case GOTSYSD_IMSG_START_PROG_APPLY_WEBCONF:
 			prog = get_helper_prog_name(imsg.hdr.type);
 			if (prog == NULL)
 				fatalx("no helper defined for imsg %d",
@@ -1017,17 +1037,38 @@ helpers_dispatch_sysconf(int fd, short event, void *arg)
 			if (gotsysd_imsg_forward(&proc->iev, &imsg, -1) == -1)
 				err = got_error_from_errno("imsg_forward");
 			break;
+		case GOTSYSD_IMSG_SYSCONF_GOTWEB_CFG:
+		case GOTSYSD_IMSG_SYSCONF_GOTWEB_ADDR:
+		case GOTSYSD_IMSG_SYSCONF_GOTWEB_ADDRS_DONE:
+		case GOTSYSD_IMSG_SYSCONF_GOTWEB_SERVER:
+		case GOTSYSD_IMSG_SYSCONF_GOTWEB_SERVERS_DONE:
+		case GOTSYSD_IMSG_SYSCONF_GOTWEB_CFG_DONE:
 		case GOTSYSD_IMSG_SYSCONF_WRITE_CONF_USERS:
 		case GOTSYSD_IMSG_SYSCONF_WRITE_CONF_USERS_DONE:
 		case GOTSYSD_IMSG_SYSCONF_WRITE_CONF_GROUP:
 		case GOTSYSD_IMSG_SYSCONF_WRITE_CONF_GROUP_MEMBERS:
 		case GOTSYSD_IMSG_SYSCONF_WRITE_CONF_GROUP_MEMBERS_DONE:
 		case GOTSYSD_IMSG_SYSCONF_WRITE_CONF_GROUPS_DONE:
+		case GOTSYSD_IMSG_SYSCONF_WEB_ACCESS_RULE:
+		case GOTSYSD_IMSG_SYSCONF_WEB_ACCESS_RULES_DONE:
 		case GOTSYSD_IMSG_SYSCONF_REPO:
+		case GOTSYSD_IMSG_SYSCONF_WEBREPO:
+		case GOTSYSD_IMSG_SYSCONF_WEBREPO_ACCESS_RULE:
+		case GOTSYSD_IMSG_SYSCONF_WEBREPO_ACCESS_RULES_DONE:
+		case GOTSYSD_IMSG_SYSCONF_WEBREPOS_DONE:
+		case GOTSYSD_IMSG_SYSCONF_WEBSITE_PATH:
+		case GOTSYSD_IMSG_SYSCONF_WEBSITE:
+		case GOTSYSD_IMSG_SYSCONF_WEBSITE_ACCESS_RULE:
+		case GOTSYSD_IMSG_SYSCONF_WEBSITE_ACCESS_RULES_DONE:
+		case GOTSYSD_IMSG_SYSCONF_WEBSITES_DONE:
 		case GOTSYSD_IMSG_SYSCONF_GLOBAL_ACCESS_RULE:
 		case GOTSYSD_IMSG_SYSCONF_GLOBAL_ACCESS_RULES_DONE:
+		case GOTSYSD_IMSG_SYSCONF_GLOBAL_MEDIA_TYPE:
+		case GOTSYSD_IMSG_SYSCONF_GLOBAL_MEDIA_TYPES_DONE:
 		case GOTSYSD_IMSG_SYSCONF_ACCESS_RULE:
 		case GOTSYSD_IMSG_SYSCONF_ACCESS_RULES_DONE:
+		case GOTSYSD_IMSG_SYSCONF_MEDIA_TYPE:
+		case GOTSYSD_IMSG_SYSCONF_MEDIA_TYPES_DONE:
 		case GOTSYSD_IMSG_SYSCONF_PROTECTED_TAG_NAMESPACES:
 		case GOTSYSD_IMSG_SYSCONF_PROTECTED_TAG_NAMESPACES_ELEM:
 		case GOTSYSD_IMSG_SYSCONF_PROTECTED_BRANCH_NAMESPACES:
@@ -1045,6 +1086,8 @@ helpers_dispatch_sysconf(int fd, short event, void *arg)
 		case GOTSYSD_IMSG_SYSCONF_NOTIFICATION_TARGET_HTTP:
 		case GOTSYSD_IMSG_SYSCONF_NOTIFICATION_TARGETS_DONE:
 		case GOTSYSD_IMSG_SYSCONF_REPOS_DONE:
+		case GOTSYSD_IMSG_SYSCONF_WEB_SERVER:
+		case GOTSYSD_IMSG_SYSCONF_WEB_SERVERS_DONE:
 			proc = find_proc(GOTSYSD_IMSG_START_PROG_WRITE_CONF,
 			    1);
 			if (proc == NULL)
@@ -1052,13 +1095,6 @@ helpers_dispatch_sysconf(int fd, short event, void *arg)
 			if (gotsysd_imsg_forward(&proc->iev, &imsg, -1) == -1)
 				err = got_error_from_errno("imsg_forward");
 			break;
-		case GOTSYSD_IMSG_SYSCONF_INSTALL_SSHD_CONFIG:
-			proc = find_proc(GOTSYSD_IMSG_START_PROG_SSHDCONFIG, 1);
-			if (proc == NULL)
-				break;
-			if (gotsysd_imsg_forward(&proc->iev, &imsg, -1) == -1)
-				err = got_error_from_errno("imsg_forward");
-			break;
 		default:
 			log_debug("unexpected imsg %d", imsg.hdr.type);
 			err = got_error(GOT_ERR_PRIVSEP_MSG);
