@@ -17,6 +17,7 @@
  */
 
 #include <sys/queue.h>
+#include <sys/tree.h>
 
 #include <errno.h>
 #include <event.h>
@@ -34,7 +35,9 @@
 #include "got_error.h"
 #include "got_reference.h"
 #include "got_object.h"
+#include "got_path.h"
 
+#include "media.h"
 #include "gotwebd.h"
 #include "log.h"
 #include "tmpl.h"
@@ -65,6 +68,7 @@ auth_shutdown(void)
 
 		config_free_access_rules(&srv->access_rules);
 		config_free_repos(&srv->repos);
+		config_free_websites(&srv->websites);
 		free(srv);
 	}
 
@@ -96,8 +100,10 @@ auth_sighdlr(int sig, short event, void *arg)
 		break;
 	case SIGCHLD:
 		break;
-	case SIGINT:
 	case SIGTERM:
+		/* continue until the parent exits */
+		break;
+	case SIGINT:
 		auth_shutdown();
 		break;
 	default:
@@ -235,7 +241,9 @@ render_error(struct request *c, const struct got_error *error)
 
 	c->t->error = error;
 
-	if (error->code == GOT_ERR_LOGIN_FAILED)
+	if (error->code == GOT_ERR_NOT_FOUND)
+		status = 404;
+	else if (error->code == GOT_ERR_LOGIN_FAILED)
 		status = 401;
 	else
 		status = 400;
@@ -380,9 +388,9 @@ logged_in:
 	}
 
 	r = tp_writef(c->tp, "Set-Cookie: gwdauth=%s;"
-	    " SameSite=Strict;%s Path=/; HttpOnly; Max-Age=%llu\r\n", token,
+	    " SameSite=Strict;%s Path=%s; HttpOnly; Max-Age=%llu\r\n", token,
 	    env->auth_config == GOTWEBD_AUTH_SECURE ? " Secure;" : "",
-	    validity);
+	    srv->gotweb_url_root, validity);
 	explicit_bzero(token, strlen(token));
 	free(token);
 	if (r == -1) {
@@ -412,24 +420,90 @@ err:
 	}
 }
 
+static void
+do_logout(struct request *c)
+{
+	const struct got_error *error = NULL;
+	struct gotwebd *env = gotwebd_env;
+	uid_t uid;
+	struct server *srv;
+	char *hostname = NULL;
+	const char *identifier = NULL;
+	struct gotweb_url url;
+
+	int r;
+
+	if (login_check_token(&uid, &hostname, c->fcgi_params.auth_cookie,
+	    auth_token_secret, sizeof(auth_token_secret),
+	    "authentication") == -1) {
+		error = got_error(GOT_ERR_LOGOUT_FAILED);
+		goto err;
+	}
+
+	/*
+	 * The www user ID represents the case where no authentication
+	 * occurred. This user must not be allowed to log in.
+	 */
+	if (uid == env->www_uid) {
+		error = got_error(GOT_ERR_LOGOUT_FAILED);
+		goto err;
+	}
+
+	c->client_uid = uid;
+	if (strcmp(hostname, c->fcgi_params.server_name) != 0) {
+		error = got_error_msg(GOT_ERR_LOGOUT_FAILED,
+		    "wrong server name in authentication cookie");
+		goto err;
+	}
+
+	srv = gotweb_get_server(c->fcgi_params.server_name);
+	if (srv == NULL) {
+		error = got_error_msg(GOT_ERR_LOGIN_FAILED,
+		    "invalid server name for logout");
+		goto err;
+	}
+
+	if (gotwebd_env->gotwebd_verbose > 0) {
+		log_info("logging out uid %u as %s for server \"%s\"",
+		    uid, identifier, hostname);
+	}
+
+	/* Ask the browser to delete the authentication cookie.  */
+	r = tp_writef(c->tp, "Set-Cookie: gwdauth=invalid;"
+	    " SameSite=Strict;%s Path=%s; HttpOnly; Max-Age=-1\r\n",
+	    env->auth_config == GOTWEBD_AUTH_SECURE ? " Secure;" : "",
+	    srv->gotweb_url_root);
+	if (r == -1) {
+		error = got_error_from_errno("tp_writef");
+		goto err;
+	}
+
+	memset(&url, 0, sizeof(url));
+	url.action = INDEX;
+	gotweb_reply(c, 307, "text/html", &url);
+	return;
+
+err:
+	free(hostname);
+	hostname = NULL;
+
+	log_warnx("%s: %s", __func__, error->msg);
+	c->t->error = error;
+	if (gotweb_reply(c, 400, "text/html", NULL) == -1)
+		return;
+	gotweb_render_page(c->tp, gotweb_render_error);
+}
+
 static const struct got_error *
 login_error_hint(struct request *c)
 {
 	struct server *srv;
-	char msg[512];
-	int ret;
 
 	srv = gotweb_get_server(c->fcgi_params.server_name);
 	if (srv == NULL || srv->login_hint_user[0] == '\0')
 		return got_error(GOT_ERR_LOGIN_FAILED);
 
-	ret = snprintf(msg, sizeof(msg),
-	    "Log in by running: ssh %s@%s \"weblogin %s\"",
-	    srv->login_hint_user, srv->name, srv->name);
-	if (ret == -1 || (size_t)ret >= sizeof(msg))
-		return got_error(GOT_ERR_LOGIN_FAILED);
-
-	return got_error_msg(GOT_ERR_LOGIN_FAILED, msg);
+	return got_error(GOT_ERR_LOGIN_HINT);
 }
 
 static void
@@ -439,10 +513,14 @@ process_request(struct request *c)
 	struct gotwebd *env = gotwebd_env;
 	uid_t uid;
 	struct server *srv;
+	struct website *site;
 	struct gotwebd_repo *repo = NULL;
 	enum gotwebd_auth_config auth_config;
+	struct gotwebd_access_rule_list *access_rules = NULL;
 	char *hostname = NULL;
 	const char *identifier = NULL;
+	char *request_path = NULL;
+	int is_repository_request = 0;
 
 	srv = gotweb_get_server(c->fcgi_params.server_name);
 	if (srv == NULL) {
@@ -451,11 +529,38 @@ process_request(struct request *c)
 		goto done;
 	}
 
+	error = gotweb_route_request(&is_repository_request, &site,
+	    &request_path, c);
+	if (error)
+		goto done;
+
+	/*
+	 * Static gotwebd assets (images, CSS, ...) are not protected
+	 * by authentication.
+	 */
+	if (is_repository_request && !got_path_is_root_dir(request_path)) {
+		forward_request(c);
+		free(request_path);
+		return;
+	}
+
+	free(request_path);
+	request_path = NULL;
+
 	auth_config = srv->auth_config;
-	if (c->fcgi_params.qs.path[0] != '\0') {
+
+	if (c->fcgi_params.qs.path[0] != '\0')
 		repo = gotweb_get_repository(srv, c->fcgi_params.qs.path);
-		if (repo)
-			auth_config = repo->auth_config;
+
+	if (repo) {
+		auth_config = repo->auth_config;
+		access_rules = &repo->access_rules;
+	} else if (site) {
+		auth_config = site->auth_config;
+		access_rules = &site->access_rules;
+
+		/* Ignore the querystring while serving web sites. */
+		fcgi_init_querystring(&c->fcgi_params.qs);
 	}
 
 	switch (auth_config) {
@@ -492,8 +597,8 @@ process_request(struct request *c)
 		goto done;
 	}
 
-	if (repo) {
-		switch (auth_check(&identifier, uid, &repo->access_rules)) {
+	if (access_rules) {
+		switch (auth_check(&identifier, uid, access_rules)) {
 		case GOTWEBD_ACCESS_DENIED:
 			error = got_error_msg(GOT_ERR_LOGIN_FAILED,
 			    "permission denied");
@@ -590,6 +695,7 @@ permitted:
 	}
 done:
 	free(hostname);
+	free(request_path);
 	if (error)
 		render_error(c, error);
 	else
@@ -717,6 +823,8 @@ auth_dispatch_sockets(int fd, short event, void *arg)
 
 			if (c->fcgi_params.qs.login[0] != '\0')
 				do_login(c);
+			else if (c->fcgi_params.qs.logout)
+				do_logout(c);
 			else
 				process_request(c);
 
@@ -878,6 +986,8 @@ auth_dispatch_main(int fd, short event, void *arg)
 	struct gotwebd		*env = gotwebd_env;
 	struct server		*srv;
 	struct gotwebd_repo	*repo;
+	struct got_pathlist_entry *pe;
+	struct website		*site;
 	ssize_t			 n;
 	int			 shut = 0;
 
@@ -909,9 +1019,18 @@ auth_dispatch_main(int fd, short event, void *arg)
 			} else {
 				srv = TAILQ_LAST(&env->servers, serverlist);
 				if (TAILQ_EMPTY(&srv->repos)) {
-					/* per-server access rule */
-					config_get_access_rule(
-					    &srv->access_rules, &imsg);
+					if (RB_EMPTY(&srv->websites)) {
+						/* per-server access rule */
+						config_get_access_rule(
+						    &srv->access_rules, &imsg);
+					} else {
+						pe = RB_MAX(got_pathlist_head,
+						    &srv->websites);
+						site = pe->data;
+						/* per-website access rule */
+						config_get_access_rule(
+						    &site->access_rules, &imsg);
+					}
 				} else {
 					/* per-repository access rule */
 					repo = TAILQ_LAST(&srv->repos,
@@ -930,6 +1049,14 @@ auth_dispatch_main(int fd, short event, void *arg)
 			srv = TAILQ_LAST(&env->servers, serverlist);
 			config_get_repository(&srv->repos, &imsg);
 			break;
+		case GOTWEBD_IMSG_CFG_WEBSITE:
+			if (TAILQ_EMPTY(&env->servers)) {
+				fatalx("%s: unexpected CFG_WEBSITE msg",
+				    __func__);
+			}
+			srv = TAILQ_LAST(&env->servers, serverlist);
+			config_get_website(&srv->websites, &imsg);
+			break;
 		case GOTWEBD_IMSG_CTL_PIPE:
 			if (env->iev_sockets == NULL)
 				recv_sockets_pipe(env, &imsg);
