@@ -19,6 +19,7 @@
 
 #include <arpa/inet.h>
 #include <sys/queue.h>
+#include <sys/tree.h>
 #include <sys/socket.h>
 #include <sys/types.h>
 #include <sys/uio.h>
@@ -40,9 +41,11 @@
 #include "got_error.h"
 #include "got_reference.h"
 #include "got_object.h"
+#include "got_path.h"
 
 #include "got_lib_poll.h"
 
+#include "media.h"
 #include "gotwebd.h"
 #include "log.h"
 #include "tmpl.h"
@@ -96,8 +99,10 @@ fcgi_sighdlr(int sig, short event, void *arg)
 		break;
 	case SIGCHLD:
 		break;
-	case SIGINT:
 	case SIGTERM:
+		/* continue until the parent exits */
+		break;
+	case SIGINT:
 		fcgi_shutdown();
 		break;
 	default:
@@ -179,14 +184,15 @@ fcgi_parse_record(struct gotwebd_fcgi_record *rec)
 }
 
 static const struct querystring_keys querystring_keys[] = {
-	{ "action",		ACTION },
-	{ "commit",		COMMIT },
-	{ "file",		RFILE },
-	{ "folder",		FOLDER },
-	{ "headref",		HEADREF },
-	{ "index_page",		INDEX_PAGE },
-	{ "path",		PATH },
-	{ "login",		LOGIN },
+	{ "action",		GOTWEBD_QS_ACTION },
+	{ "commit",		GOTWEBD_QS_COMMIT },
+	{ "file",		GOTWEBD_QS_RFILE },
+	{ "folder",		GOTWEBD_QS_FOLDER },
+	{ "headref",		GOTWEBD_QS_HEADREF },
+	{ "index_page",	GOTWEBD_QS_INDEX_PAGE },
+	{ "path",		GOTWEBD_QS_PATH },
+	{ "login",		GOTWEBD_QS_LOGIN },
+	{ "logout",		GOTWEBD_QS_LOGOUT },
 };
 
 static const struct action_keys action_keys[] = {
@@ -294,7 +300,7 @@ assign_querystring(struct querystring *qs, char *key, char *value)
 			continue;
 
 		switch (querystring_keys[el_cnt].element) {
-		case ACTION:
+		case GOTWEBD_QS_ACTION:
 			for (a_cnt = 0; a_cnt < nitems(action_keys); a_cnt++) {
 				if (strcmp(value, action_keys[a_cnt].name) != 0)
 					continue;
@@ -304,7 +310,7 @@ assign_querystring(struct querystring *qs, char *key, char *value)
 			if (a_cnt == nitems(action_keys))
 				qs->action = ERR;
 			break;
-		case COMMIT:
+		case GOTWEBD_QS_COMMIT:
 			if (strlcpy(qs->commit, value, sizeof(qs->commit)) >=
 			    sizeof(qs->commit)) {
 				error = got_error_msg(GOT_ERR_NO_SPACE,
@@ -312,7 +318,7 @@ assign_querystring(struct querystring *qs, char *key, char *value)
 				goto done;
 			}
 			break;
-		case RFILE:
+		case GOTWEBD_QS_RFILE:
 			error = validate_path(value);
 			if (error)
 				goto done;
@@ -323,7 +329,7 @@ assign_querystring(struct querystring *qs, char *key, char *value)
 				goto done;
 			}
 			break;
-		case FOLDER:
+		case GOTWEBD_QS_FOLDER:
 			error = validate_path(value);
 			if (error)
 				goto done;
@@ -335,7 +341,7 @@ assign_querystring(struct querystring *qs, char *key, char *value)
 				goto done;
 			}
 			break;
-		case HEADREF:
+		case GOTWEBD_QS_HEADREF:
 			error = validate_path(value);
 			if (error)
 				goto done;
@@ -346,7 +352,7 @@ assign_querystring(struct querystring *qs, char *key, char *value)
 				goto done;
 			}
 			break;
-		case INDEX_PAGE:
+		case GOTWEBD_QS_INDEX_PAGE:
 			if (*value == '\0')
 				break;
 			qs->index_page = strtonum(value, 0, INT_MAX, &errstr);
@@ -356,7 +362,7 @@ assign_querystring(struct querystring *qs, char *key, char *value)
 				goto done;
 			}
 			break;
-		case PATH:
+		case GOTWEBD_QS_PATH:
 			error = validate_path(value);
 			if (error)
 				goto done;
@@ -367,7 +373,7 @@ assign_querystring(struct querystring *qs, char *key, char *value)
 				goto done;
 			}
 			break;
-		case LOGIN:
+		case GOTWEBD_QS_LOGIN:
 			if (strlcpy(qs->login, value, sizeof(qs->login)) >=
 			    sizeof(qs->login)) {
 				error = got_error_msg(GOT_ERR_NO_SPACE,
@@ -375,6 +381,11 @@ assign_querystring(struct querystring *qs, char *key, char *value)
 				goto done;
 			}
 			break;
+		case GOTWEBD_QS_LOGOUT:
+			if (strcmp(value, "1") != 0)
+				break;
+			qs->logout = 1;
+			break;
  		}
 
 		/* entry found */
