@@ -50,13 +50,14 @@
 #include "got_blame.h"
 #include "got_privsep.h"
 
+#include "media.h"
 #include "gotwebd.h"
 #include "log.h"
 #include "tmpl.h"
 
 static int gotweb_render_index(struct template *);
 static const struct got_error *gotweb_load_got_path(struct repo_dir **,
-    const char *, struct request *);
+    const char *, struct request *, struct website *);
 static const struct got_error *gotweb_load_file(char **, const char *,
     const char *, int);
 static const struct got_error *gotweb_get_repo_description(char **,
@@ -83,7 +84,7 @@ gotweb_reply(struct request *c, int status, const char *ctype,
 	}
 
 	csp = "Content-Security-Policy: default-src 'self'; "
-	    "script-src 'none'; object-src 'none';\r\n";
+	    "script-src 'self'; object-src 'none';\r\n";
 	if (tp_writes(c->tp, csp) == -1)
 		return -1;
 
@@ -217,11 +218,507 @@ recv_request(struct imsg *imsg)
 	return c;
 }
 
+void
+gotweb_log_request(struct request *c)
+{
+	struct gotwebd_fcgi_params *p = &c->fcgi_params;
+	struct querystring *qs = &p->qs;
+	char *server_name = NULL;
+	char *document_uri = NULL;
+	const char *action_name = NULL;
+
+	if (gotwebd_env->gotwebd_verbose == 0)
+		return;
+
+	if (p->server_name[0] &&
+	    stravis(&server_name, p->server_name, VIS_SAFE) == -1) {
+		log_warn("stravis");
+		server_name = NULL;
+	}
+
+	if (p->document_uri[0] &&
+	    stravis(&document_uri, p->document_uri, VIS_SAFE) == -1) {
+		log_warn("stravis");
+		document_uri = NULL;
+	}
+
+	action_name = gotweb_action_name(qs->action);
+	log_info("processing request: server='%s' action='%s' "
+	    "commit='%s', file='%s', folder='%s', headref='%s' "
+	    "index_page=%d path='%s' document_uri='%s'",
+	    server_name ? server_name : "",
+	    action_name ? action_name : "",
+	    qs->commit,
+	    qs->file,
+	    qs->folder,
+	    qs->headref,
+	    qs->index_page,
+	    qs->path,
+	    document_uri ? document_uri : "");
+
+	free(server_name);
+	free(document_uri);
+}
+
+const struct got_error *
+gotweb_serve_htdocs(struct request *c, const char *request_path)
+{
+	const struct got_error *error = NULL;
+	struct server *srv = c->srv;;
+	struct media_type *m;
+	char *ondisk_path = NULL;
+	int n, fd = -1;
+	char mime_type[MEDIA_STRMAX] = "application/octet-stream";
+
+	if (asprintf(&ondisk_path, "%s/%s/%s", gotwebd_env->httpd_chroot,
+	    srv->htdocs_path, request_path) == -1) {
+		error = got_error_from_errno("asprintf");
+		goto done;
+	}
+
+	fd = open(ondisk_path, O_RDONLY);
+	if (fd == -1) {
+		if (errno == ENOENT || errno == ENOTDIR)
+			error = got_error(GOT_ERR_NOT_FOUND);
+		else
+			error = got_error_from_errno_fmt("open: %s");
+		goto done;
+	}
+
+	m = media_find(&gotwebd_env->mediatypes, ondisk_path);
+	if (m != NULL) {
+		n = snprintf(mime_type, sizeof(mime_type), "%s/%s",
+		    m->media_type, m->media_subtype);
+		if (n < 0 || (size_t)n >= sizeof(mime_type)) {
+			error = got_error(GOT_ERR_RANGE);
+			goto done;
+		}
+	}
+
+	if (gotweb_reply(c, 200, mime_type, NULL) == -1) {
+		error = got_error(GOT_ERR_IO);
+		goto done;
+	}
+
+	if (template_flush(c->tp) == -1) {
+		error = got_error(GOT_ERR_IO);
+		goto done;
+	}
+
+	for (;;) {
+		uint8_t buf[BUF];
+		ssize_t r;
+
+		r = read(fd, buf, sizeof(buf));
+		if (r == -1) {
+			error = got_error_from_errno("read");
+			goto done;
+		}
+		if (r == 0)
+			break;
+	
+		if (fcgi_write(c, buf, r) == -1) {
+			error = got_error(GOT_ERR_IO);
+			goto done;
+		}
+	}
+done:
+	if (fd != -1 && close(fd) == -1 && error == NULL)
+		error = got_error_from_errno("close");
+	free(ondisk_path);
+	return error;
+}
+
+static const struct got_error *
+serve_blob(int *response_code, struct request *c, struct got_repository *repo,
+    struct got_object_id *obj_id, const char *basename)
+{
+	const struct got_error *error = NULL;
+	struct got_blob_object *blob = NULL;
+	struct media_type *m;
+	int binary, n;
+	char mime_type[MEDIA_STRMAX] = "application/octet-stream";
+
+	c->t->fd = dup(c->priv_fd[BLOB_FD_1]);
+	if (c->t->fd == -1) {
+		error = got_error_from_errno("dup");
+		goto done;
+	}
+
+	error = got_object_open_as_blob(&blob, repo, obj_id, BUF, c->t->fd);
+	if (error)
+		goto done;
+
+	error = got_object_blob_is_binary(&binary, blob);
+	if (error)
+		goto done;
+
+	if (binary) {
+		if (gotweb_reply_file(c, mime_type, basename, NULL) == -1) {
+			error = got_error(GOT_ERR_IO);
+			*response_code = 500;
+			goto done;
+		}
+	} else {
+		m = media_find(&gotwebd_env->mediatypes, basename);
+		if (m != NULL) {
+			n = snprintf(mime_type, sizeof(mime_type), "%s/%s",
+			    m->media_type, m->media_subtype);
+			if (n < 0 || (size_t)n >= sizeof(mime_type)) {
+				error = got_error_msg(GOT_ERR_RANGE,
+				    "media type snprintf");
+				goto done;
+			}
+		}
+
+		if (gotweb_reply(c, 200, mime_type, NULL) == -1) {
+			error = got_error(GOT_ERR_IO);
+			*response_code = 500;
+			goto done;
+		}
+	}
+
+	if (template_flush(c->tp) == -1) {
+		error = got_error(GOT_ERR_IO);
+		*response_code = 500;
+		goto done;
+	}
+
+	for (;;) {
+		const uint8_t *buf;
+		size_t len;
+
+		error = got_object_blob_read_block(&len, blob);
+		if (error)
+			goto done;
+		if (len == 0)
+			break;
+
+		buf = got_object_blob_get_read_buf(blob);
+		if (fcgi_write(c, buf, len) == -1) {
+			error = got_error(GOT_ERR_IO);
+			*response_code = 500;
+			goto done;
+		}
+	}
+
+done:
+	if (blob)
+		got_object_blob_close(blob);
+	return error;
+}
+
+static int
+gotweb_serve_website(struct request *c, struct website *site,
+    struct repo_dir *repo_dir, const char *request_path)
+{
+	const struct got_error *error = NULL;
+	struct transport *t = c->t;
+	struct gotwebd_fcgi_params *p = &c->fcgi_params;
+	struct got_repository *repo = t->repo;
+	char *refname = NULL;
+	struct got_reference *ref = NULL;
+	struct got_object_id *id = NULL, *obj_id = NULL;
+	struct got_commit_object *commit = NULL;
+	struct got_tree_object *tree = NULL;
+	char *basename = NULL;
+	int response_code = 404;
+	char *in_repo_child = NULL, *in_repo_path = NULL;
+	int obj_type;
+
+	if (got_path_is_root_dir(request_path)) {
+		in_repo_child = strdup(request_path);
+		if (in_repo_child == NULL) {
+			error = got_error_from_errno("strdup");
+			goto done;
+		}
+	} else {
+		if (got_path_is_root_dir(site->url_path)) {
+			in_repo_child = strdup(request_path);
+			if (in_repo_child == NULL) {
+				error = got_error_from_errno("strdup");
+				goto done;
+			}
+		} else if (got_path_cmp(site->url_path, request_path,
+		    strlen(site->url_path), strlen(request_path)) == 0) {
+			in_repo_child = strdup("/");
+			if (in_repo_child == NULL) {
+				error = got_error_from_errno("strdup");
+				goto done;
+			}
+		} else {
+			error = got_path_skip_common_ancestor(&in_repo_child,
+			   site->url_path, request_path);
+			if (error)
+				goto done;
+		}
+	}
+
+	if (site->path[0] != '\0') {
+		char *s = in_repo_child;
+
+		while (*s == '/')
+			s++;
+
+		if (asprintf(&in_repo_path, "%s/%s", site->path, s) == -1) {
+			error = got_error_from_errno("asprintf");
+			goto done;
+		}
+
+		got_path_strip_trailing_slashes(in_repo_path);
+	} else {
+		in_repo_path = in_repo_child;
+		in_repo_child = NULL;
+	}
+
+	if (site->branch_name[0] != 0) {
+		const char *branch = site->branch_name;
+
+		if (strncmp("refs/heads/", branch, 11) != 0) {
+			if (asprintf(&refname, "refs/heads/%s", branch) == -1) {
+				error = got_error_from_errno("asprintf");
+				goto done;
+			}
+		} else {
+			refname = strdup(branch);
+			if (refname == NULL) {
+				error = got_error_from_errno("strdup");
+				goto done;
+			}
+		}
+
+	} else {
+		refname = strdup(GOT_REF_HEAD);
+		if (refname == NULL) {
+			error = got_error_from_errno("strdup");
+			goto done;
+		}
+	}
+
+	error = got_ref_open(&ref, repo, refname, 0);
+	if (error)
+		goto done;
+
+	error = got_ref_resolve(&id, repo, ref);
+	if (error)
+		goto done;
+
+	error = got_object_open_as_commit(&commit, repo, id);
+	if (error)
+		goto done;
+
+	error = got_object_id_by_path(&obj_id, repo, commit, in_repo_path);
+	if (error)
+		goto done;
+
+	error = got_object_get_type(&obj_type, repo, obj_id);
+	if (error)
+		goto done;
+
+	switch (obj_type) {
+	case GOT_OBJ_TYPE_BLOB:
+		error = got_path_basename(&basename, in_repo_path);
+		if (error)
+			goto done;
+		error = serve_blob(&response_code, c, repo, obj_id,
+		    basename);
+		if (error)
+			goto done;
+		break;
+	case GOT_OBJ_TYPE_TREE: {
+		struct got_tree_entry *te;
+		int nentries, i;
+		const char *name;
+		mode_t mode;
+
+		error = got_object_open_as_tree(&tree, repo, obj_id);
+		if (error)
+			goto done;
+		nentries = got_object_tree_get_nentries(tree);
+
+		for (i = 0; i < nentries; i++) {
+			struct gotweb_url url = {
+				.index_page = -1,
+				.action = -1,
+			};
+
+			te = got_object_tree_get_entry(tree, i);
+
+			name = got_tree_entry_get_name(te);
+			mode = got_tree_entry_get_mode(te);
+			if (!S_ISREG(mode) ||
+			    strcasecmp(name, "index.html") != 0)
+				continue;
+
+			/* XXX gotweb_reply uses request struct field */
+			got_path_strip_trailing_slashes(p->document_uri);
+			if (strlcat(p->document_uri, "/index.html",
+			    sizeof(p->document_uri)) >=
+			    sizeof(p->document_uri)) {
+				error = got_error(GOT_ERR_NO_SPACE);
+				goto done;
+			}
+
+			if (gotweb_reply(c, 302, NULL, &url) == -1) {
+				error = got_error(GOT_ERR_IO);
+				goto done;
+			}
+			break;
+		}
+		break;
+	}
+	default:
+		error = got_error(GOT_ERR_NOT_FOUND);
+		goto done;
+	}
+done:
+	free(in_repo_child);
+	free(in_repo_path);
+	free(refname);
+	free(basename);
+	free(id);
+	free(obj_id);
+	if (ref)
+		got_ref_close(ref);
+	if (commit)
+		got_object_commit_close(commit);
+	if (tree)
+		got_object_tree_close(tree);
+
+	if (error) {
+		char *safe_path = NULL;
+
+		if (stravis(&safe_path, request_path, VIS_SAFE) == -1) {
+			log_warn("stravis");
+			safe_path = NULL;
+		}
+		log_warnx("%s: %s: %d: %s", __func__,
+		    safe_path ? safe_path : "?", response_code, error->msg);
+		free(safe_path);
+
+		if (response_code == 404)
+			c->t->error = got_error(GOT_ERR_NOT_FOUND);
+		else
+			c->t->error = error;
+
+		if (gotweb_reply(c, response_code, "text/html", NULL) == -1)
+			return -1;
+		return gotweb_render_page(c->tp, gotweb_render_error);
+	}
+
+	return 0;
+}
+
+const struct got_error *
+gotweb_route_request(int *is_repository_request, struct website **site,
+	char **request_path, struct request *c)
+{
+	const struct got_error *error = NULL;
+	struct gotwebd_fcgi_params *p = &c->fcgi_params;
+	struct server *srv = c->srv;;
+	char *child_path = NULL;
+
+	*is_repository_request = 0;
+	*site = NULL;
+	*request_path = NULL;
+
+	if (got_path_cmp(srv->full_repos_url_path, p->document_uri,
+	    strlen(srv->full_repos_url_path),
+	    strlen(p->document_uri)) == 0) {
+		/* 
+		 * Requesting / in repository url path space.
+		 * We will be rendering Git repository data.
+		 */
+		*request_path = strdup("/");
+		if (*request_path == NULL) {
+			error = got_error_from_errno("strdup");
+			goto done;
+		}
+		*is_repository_request = 1;
+	} else if (got_path_is_child(p->document_uri,
+	    srv->full_repos_url_path, strlen(srv->full_repos_url_path))) {
+		/*
+		 * Requesting something within repository url path space.
+		 * We will be returning a static asset for a repository page.
+		 */
+		error = got_path_skip_common_ancestor(&child_path,
+		    srv->full_repos_url_path, p->document_uri);
+		if (error)
+			goto done;
+		*is_repository_request = 1;
+
+		if (asprintf(request_path, "/%s", child_path) == -1) {
+			error = got_error_from_errno("asprintf");
+			goto done;
+		}
+	} else if (got_path_is_child(p->document_uri,
+	    srv->gotweb_url_root, strlen(srv->gotweb_url_root))) {
+		/*
+		 * Requesting something outside repository url path space.
+		 * This will result in a 404 error unless the request is
+		 * matched by a website.
+		 */
+		if (got_path_is_root_dir(p->document_uri)) {
+			*request_path = strdup("/");
+			if (*request_path == NULL) {
+				error = got_error_from_errno("strdup");
+				goto done;
+			}
+		} else {
+			error = got_path_skip_common_ancestor(&child_path,
+			    srv->gotweb_url_root, p->document_uri);
+			if (error)
+				goto done;
+			if (asprintf(request_path, "/%s", child_path) == -1) {
+				error = got_error_from_errno("asprintf");
+				goto done;
+			}
+		}
+	} else {
+		/*
+		 * Requesting something outside gotweb url path space.
+		 * This will result in a 404 error.
+		 */
+		*request_path = strdup(p->document_uri);
+		if (*request_path == NULL) {
+			error = got_error_from_errno("strdup");
+			goto done;
+		}
+	}
+
+	if (!got_path_is_root_dir(*request_path))
+		got_path_strip_trailing_slashes(*request_path);
+
+	*site = gotweb_get_website(srv, *request_path);
+
+	/*
+	 * If the target of the request is ambiguous, the longer path wins.
+	 */
+	if (*is_repository_request && *site) {
+		if (got_path_cmp(srv->repos_url_path, (*site)->url_path,
+		    strlen(srv->repos_url_path),
+		    strlen((*site)->url_path)) <= 0)
+			*is_repository_request = 0;
+		else
+			*site = NULL;
+	}
+done:
+	free(child_path);
+	if (error) {
+		free(*request_path);
+		*request_path = NULL;
+		*site = NULL;
+	}
+
+	return error;
+}
+
 int
 gotweb_process_request(struct request *c)
 {
 	const struct got_error *error = NULL;
 	struct server *srv = c->srv;;
+	struct website *site;
 	struct querystring *qs = NULL;
 	struct repo_dir *repo_dir = NULL;
 	struct repo_commit *commit;
@@ -229,47 +726,44 @@ gotweb_process_request(struct request *c)
 	const uint8_t *buf;
 	size_t len;
 	int r, binary = 0;
+	char *request_path = NULL;
+	int is_repository_request = 0;
 
 	/* querystring */
 	qs = &c->fcgi_params.qs;
 	c->t->qs = qs;
 
-	/* Log the request. */
-	if (gotwebd_env->gotwebd_verbose > 0) {
-		struct gotwebd_fcgi_params *p = &c->fcgi_params;
-		char *server_name = NULL;
-		char *document_uri = NULL;
-		const char *action_name = NULL;
+	gotweb_log_request(c);
 
-		if (p->server_name[0] &&
-		    stravis(&server_name, p->server_name, VIS_SAFE) == -1) {
-			log_warn("stravis");
-			server_name = NULL;
-		}
+	error = gotweb_route_request(&is_repository_request, &site,
+	    &request_path, c);
+	if (error)
+		goto err;
 
-		if (p->document_uri[0] &&
-		    stravis(&document_uri, p->document_uri, VIS_SAFE) == -1) {
-			log_warn("stravis");
-			document_uri = NULL;
-		}
+	if (is_repository_request && !got_path_is_root_dir(request_path)) {
+		error = gotweb_serve_htdocs(c, request_path);
+		if (error)
+			goto err;
 
-		action_name = gotweb_action_name(qs->action);
-		log_info("processing request: server='%s' action='%s' "
-		    "commit='%s', file='%s', folder='%s', headref='%s' "
-		    "index_page=%d path='%s' document_uri='%s'",
-		    server_name ? server_name : "",
-		    action_name ? action_name : "",
-		    qs->commit,
-		    qs->file,
-		    qs->folder,
-		    qs->headref,
-		    qs->index_page,
-		    qs->path,
-		    document_uri ? document_uri : "");
-		free(server_name);
-		free(document_uri);
+		free(request_path);
+		return 0;
 	}
 
+	if (site) {
+		error = gotweb_load_got_path(&repo_dir, site->repo_name, c,
+		    site);
+		c->t->repo_dir = repo_dir;
+		if (error)
+			goto err;
+
+		r = gotweb_serve_website(c, site, repo_dir, request_path);
+		free(request_path);
+		return r;
+	}
+
+	free(request_path);
+	request_path = NULL;
+
 	/*
 	 * certain actions require a commit id in the querystring. this stops
 	 * bad actors from exploiting this by manually manipulating the
@@ -291,7 +785,7 @@ gotweb_process_request(struct request *c)
 			goto err;
 		}
 
-		error = gotweb_load_got_path(&repo_dir, qs->path, c);
+		error = gotweb_load_got_path(&repo_dir, qs->path, c, NULL);
 		c->t->repo_dir = repo_dir;
 		if (error)
 			goto err;
@@ -500,9 +994,16 @@ gotweb_process_request(struct request *c)
 	}
 
 err:
+	free(request_path);
 	c->t->error = error;
-	if (gotweb_reply(c, 400, "text/html", NULL) == -1)
-		return -1;
+	if (error->code == GOT_ERR_NOT_FOUND) {
+		if (gotweb_reply(c, 404, "text/html", NULL) == -1)
+			return -1;
+	} else {
+		if (gotweb_reply(c, 400, "text/html", NULL) == -1)
+			return -1;
+	}
+
 	return gotweb_render_page(c->tp, gotweb_render_error);
 }
 
@@ -521,6 +1022,36 @@ gotweb_get_server(const char *server_name)
 	return TAILQ_FIRST(&gotwebd_env->servers);
 };
 
+struct website *
+gotweb_get_website(struct server *srv, const char *url_path)
+{
+	struct got_pathlist_entry *pe;
+	size_t url_path_len = strlen(url_path);
+	struct website *site = NULL;
+
+	if (RB_EMPTY(&srv->websites))
+		return NULL;
+
+	/*
+	 * Parent paths are sorted before children so we can match the
+	 * provided URL path against the most specific web site URL path
+	 * by walking the list backwards.
+	 */
+	pe = RB_MAX(got_pathlist_head, &srv->websites);
+	while (pe) {
+		if (got_path_cmp(url_path, pe->path,
+		    url_path_len, pe->path_len) == 0 ||
+		    got_path_is_child(url_path, pe->path, pe->path_len)) {
+			site = pe->data;
+			break;
+		}
+
+		pe = RB_PREV(got_pathlist_head, &srv->websites, pe);
+	}
+
+	return site;
+}
+
 const struct got_error *
 gotweb_init_transport(struct transport **t)
 {
@@ -593,11 +1124,15 @@ gotweb_free_repo_commit(struct repo_commit *rc)
 static void
 gotweb_free_repo_dir(struct repo_dir *repo_dir)
 {
+	size_t i;
+
 	if (repo_dir != NULL) {
 		free(repo_dir->name);
 		free(repo_dir->owner);
 		free(repo_dir->description);
 		free(repo_dir->url);
+		for (i = 0; i < nitems(repo_dir->sshfp); i++)
+			free(repo_dir->sshfp[i]);
 		free(repo_dir->path);
 	}
 	free(repo_dir);
@@ -712,7 +1247,7 @@ gotweb_render_index(struct template *tp)
 		}
 
 		error = gotweb_load_got_path(&repo_dir, sd_dent[d_i]->d_name,
-		    c);
+		    c, NULL);
 		if (error) {
 			if (error->code != GOT_ERR_NOT_GIT_REPO)
 				log_warnx("%s: %s: %s", __func__,
@@ -934,6 +1469,12 @@ gotweb_render_url(struct request *c, struct gotweb_url *url)
 		sep = "&";
 	}
 
+	if (url->logout == 1) {
+		if (tp_writef(c->tp, "%slogout=1", sep) == -1)
+			return -1;
+		sep = "&";
+	}
+
 	return 0;
 }
 
@@ -996,7 +1537,7 @@ auth_check(struct request *c, struct gotwebd_access_rule_list *rules)
 
 	/*
 	 * We cannot access /etc/passwd in this process so we cannot
-	 * verify the client's user ID outselves here.
+	 * verify the client's user ID ourselves here.
 	 * Match rules against the access identifier which has already
 	 * passed authentication in the auth process.
 	 */
@@ -1010,7 +1551,7 @@ auth_check(struct request *c, struct gotwebd_access_rule_list *rules)
 
 static const struct got_error *
 gotweb_load_got_path(struct repo_dir **rp, const char *dir,
-    struct request *c)
+    struct request *c, struct website *site)
 {
 	const struct got_error *error = NULL;
 	struct gotwebd *env = gotwebd_env;
@@ -1021,6 +1562,7 @@ gotweb_load_got_path(struct repo_dir **rp, const char *dir,
 	char *dir_test;
 	struct gotwebd_repo *repo;
 	enum gotwebd_auth_config auth_config = 0;
+	struct gotwebd_access_rule_list *access_rules = NULL;
 	enum gotwebd_access access = GOTWEBD_ACCESS_DENIED;
 	int repo_is_hidden = 0;
 
@@ -1067,16 +1609,23 @@ gotweb_load_got_path(struct repo_dir **rp, const char *dir,
 		goto err;
 
 	repo = gotweb_get_repository(srv, repo_dir->name);
-	if (repo) {
-		repo_is_hidden = repo->hidden;
-		auth_config = repo->auth_config;
+	if (repo || site) {
+		if (site) {
+			auth_config = site->auth_config;
+			access_rules = &site->access_rules;
+		} else {
+			repo_is_hidden = repo->hidden;
+			auth_config = repo->auth_config;
+			access_rules = &repo->access_rules;
+		}
+
 		switch (auth_config) {
 		case GOTWEBD_AUTH_DISABLED:
 			access = GOTWEBD_ACCESS_PERMITTED;
 			break;
 		case GOTWEBD_AUTH_SECURE:
 		case GOTWEBD_AUTH_INSECURE:
-			access = auth_check(c, &repo->access_rules);
+			access = auth_check(c, access_rules);
 			if (access == GOTWEBD_ACCESS_NO_MATCH)
 				access = auth_check(c, &srv->access_rules);
 			if (access == GOTWEBD_ACCESS_NO_MATCH)
@@ -1112,25 +1661,37 @@ gotweb_load_got_path(struct repo_dir **rp, const char *dir,
 		goto err;
 	}
 
-	if (repo_is_hidden) {
-		error = got_error_path(repo_dir->name, GOT_ERR_NOT_GIT_REPO);
-		goto err;
-	}
+	if (site == NULL) {
+		if (repo_is_hidden) {
+			error = got_error_path(repo_dir->name,
+			    GOT_ERR_NOT_GIT_REPO);
+			goto err;
+		}
 
-	if (srv->respect_exportok &&
-	    faccessat(dirfd(dt), "git-daemon-export-ok", F_OK, 0) == -1) {
-		error = got_error_path(repo_dir->name, GOT_ERR_NOT_GIT_REPO);
-		goto err;
+		if (srv->respect_exportok && faccessat(dirfd(dt),
+		    "git-daemon-export-ok", F_OK, 0) == -1) {
+			error = got_error_path(repo_dir->name,
+			    GOT_ERR_NOT_GIT_REPO);
+			goto err;
+		}
 	}
 
 	error = got_repo_open(&t->repo, repo_dir->path, NULL,
 	    gotwebd_env->pack_fds);
 	if (error)
 		goto err;
-	error = gotweb_get_repo_description(&repo_dir->description, srv,
-	    repo_dir->path, dirfd(dt));
-	if (error)
-		goto err;
+	if (repo && repo->description[0] != '\0') {
+		repo_dir->description = strdup(repo->description);
+		if (repo_dir->description == NULL) {
+			error = got_error_from_errno("strdup");
+			goto err;
+		}
+	} else {
+		error = gotweb_get_repo_description(&repo_dir->description,
+		    srv, repo_dir->path, dirfd(dt));
+		if (error)
+			goto err;
+	}
 	if (srv->show_repo_owner) {
 		error = gotweb_load_file(&repo_dir->owner, repo_dir->path,
 		    "owner", dirfd(dt));
@@ -1147,8 +1708,33 @@ gotweb_load_got_path(struct repo_dir **rp, const char *dir,
 		if (error)
 			goto err;
 	}
-	error = gotweb_get_clone_url(&repo_dir->url, srv, repo_dir->path,
-	    dirfd(dt));
+	if (srv->show_repo_cloneurl && repo && repo->clone_url[0] != '\0') {
+		repo_dir->url = strdup(repo->clone_url);
+		if (repo_dir->url == NULL) {
+			error = got_error_from_errno("strdup");
+			goto err;
+		}
+	} else {
+		error = gotweb_get_clone_url(&repo_dir->url, srv,
+		    repo_dir->path, dirfd(dt));
+	}
+
+	if (srv->show_repo_cloneurl && repo) {
+		size_t i;
+
+		for (i = 0; i < nitems(repo->clone_url_hostkey); i++) {
+			if (repo->clone_url_hostkey[i][0] == '\0')
+				continue;
+
+			repo_dir->sshfp[i] = strdup(repo->clone_url_hostkey[i]);
+			if (repo_dir->sshfp[i] == NULL) {
+				error = got_error_from_errno("strdup");
+				goto err;
+			}
+		}
+
+	}
+
 err:
 	free(dir_test);
 	if (dt != NULL && closedir(dt) == EOF && error == NULL)
@@ -1278,6 +1864,8 @@ gotweb_shutdown(void)
 	imsgbuf_clear(&gotwebd_env->iev_auth->ibuf);
 	free(gotwebd_env->iev_auth);
 
+	media_purge(&gotwebd_env->mediatypes);
+
 	config_free_access_rules(&gotwebd_env->access_rules);
 
 	while (!TAILQ_EMPTY(&gotwebd_env->servers)) {
@@ -1288,6 +1876,7 @@ gotweb_shutdown(void)
 
 		config_free_access_rules(&srv->access_rules);
 		config_free_repos(&srv->repos);
+		config_free_websites(&srv->websites);
 		free(srv);
 	}
 
@@ -1327,8 +1916,10 @@ gotweb_sighdlr(int sig, short event, void *arg)
 		break;
 	case SIGCHLD:
 		break;
-	case SIGINT:
 	case SIGTERM:
+		/* continue until the parent exits */
+		break;
+	case SIGINT:
 		gotweb_shutdown();
 		break;
 	default:
@@ -1336,6 +1927,30 @@ gotweb_sighdlr(int sig, short event, void *arg)
 	}
 }
 
+static void
+unveil_htdocs_path(const char *htdocs_path)
+{
+	struct gotwebd *env = gotwebd_env;
+	char path[PATH_MAX];
+	int ret;
+
+	while (htdocs_path[0] == '/')
+		htdocs_path++;
+
+	ret = snprintf(path, sizeof(path), "%s/%s",
+	    env->httpd_chroot, htdocs_path);
+	if (ret == -1)
+		fatal("snprintf");
+	if ((size_t)ret >= sizeof(path)) {
+		fatalx("htdocs path too long, exceeds %zd bytes: %s",
+		    sizeof(path) - strlen(env->httpd_chroot) - 1,
+		    htdocs_path);
+	}
+
+	if (unveil(path, "r") == -1)
+		fatal("unveil %s", path);
+}
+
 static void
 gotweb_launch(struct gotwebd *env)
 {
@@ -1350,9 +1965,15 @@ gotweb_launch(struct gotwebd *env)
 		fatal("pledge");
 #endif
 
+	unveil_htdocs_path(env->htdocs_path);
+
 	TAILQ_FOREACH(srv, &gotwebd_env->servers, entry) {
 		if (unveil(srv->repos_path, "r") == -1)
 			fatal("unveil %s", srv->repos_path);
+
+		if (got_path_cmp(env->htdocs_path, srv->htdocs_path, 
+		    strlen(env->htdocs_path), strlen(srv->htdocs_path)) != 0)
+			unveil_htdocs_path(srv->htdocs_path);
 	}
 
 	error = got_privsep_unveil_exec_helpers();
@@ -1510,6 +2131,9 @@ gotweb_dispatch_main(int fd, short event, void *arg)
 				}
 			}
 			break;
+		case GOTWEBD_IMSG_CFG_MEDIA_TYPE:
+			config_getmediatype(env, &imsg);
+			break;
 		case GOTWEBD_IMSG_CFG_SRV:
 			config_getserver(env, &imsg);
 			break;
@@ -1519,6 +2143,14 @@ gotweb_dispatch_main(int fd, short event, void *arg)
 			srv = TAILQ_LAST(&env->servers, serverlist);
 			config_get_repository(&srv->repos, &imsg);
 			break;
+		case GOTWEBD_IMSG_CFG_WEBSITE:
+			if (TAILQ_EMPTY(&env->servers)) {
+				fatalx("%s: unexpected CFG_WEBSITE msg",
+				    __func__);
+			}
+			srv = TAILQ_LAST(&env->servers, serverlist);
+			config_get_website(&srv->websites, &imsg);
+			break;
 		case GOTWEBD_IMSG_CFG_FD:
 			config_getfd(env, &imsg);
 			break;
