@@ -20,6 +20,7 @@
 #include <sys/param.h>
 #include <sys/ioctl.h>
 #include <sys/queue.h>
+#include <sys/tree.h>
 #include <sys/wait.h>
 #include <sys/uio.h>
 #include <sys/resource.h>
@@ -54,7 +55,9 @@
 
 #include "got_reference.h"
 #include "got_object.h"
+#include "got_path.h"
 
+#include "media.h"
 #include "gotwebd.h"
 #include "log.h"
 #include "tmpl.h"
@@ -63,6 +66,7 @@
 #define MAXIMUM(a, b)	(((a) > (b)) ? (a) : (b))
 
 static volatile int client_cnt;
+static volatile int stopping;
 
 static struct timeval	timeout = { TIMEOUT_DEFAULT, 0 };
 
@@ -145,7 +149,8 @@ cleanup_request(struct request *c)
 
 	del_request(c);
 
-	event_add(&c->sock->ev, NULL);
+	if (!stopping)
+		event_add(&c->sock->ev, NULL);
 
 	if (evtimer_initialized(&c->tmo))
 		evtimer_del(&c->tmo);
@@ -155,6 +160,9 @@ cleanup_request(struct request *c)
 		close(c->fd);
 	free(c->buf);
 	free(c);
+
+	if (stopping && client_cnt <= 0)
+		sockets_shutdown();
 }
 
 static void
@@ -555,6 +563,7 @@ process_request(struct request *c)
 	}
 
 	c->fd = -1;
+	event_del(&c->ev);
 	c->client_status = CLIENT_REQUEST;
 	env->worker_load[c->worker_idx]++;
 	return 0;
@@ -631,6 +640,9 @@ recv_parsed_params(struct imsg *imsg)
 		goto fail;
 	}
 
+	if (params.qs.logout)
+		p->qs.logout = 1;
+
 	if (params.document_uri[0] != '\0' &&
 	    strlcpy(p->document_uri, params.document_uri,
 	    sizeof(p->document_uri)) >= sizeof(p->document_uri)) {
@@ -752,6 +764,32 @@ recv_fcgi_pipe(struct gotwebd *env, struct imsg *imsg)
 	env->iev_fcgi = iev;
 }
 
+static void
+sockets_stop(void)
+{
+	struct socket *sock;
+
+	stopping = 1;
+
+	/*
+	 * Deactivate sockets but to do not free them.
+	 * Client connections maintain a pointer to them via c->sock.
+	 */
+	TAILQ_FOREACH(sock, &gotwebd_env->sockets, entry) {
+		if (sock->fd != -1) {
+			close(sock->fd);
+			sock->fd = -1;
+		}
+		if (event_initialized(&sock->ev))
+			event_del(&sock->ev);
+		if (evtimer_initialized(&sock->ev))
+			evtimer_del(&sock->pause);
+	}
+
+	if (client_cnt == 0)
+		sockets_shutdown();
+}
+
 static void
 sockets_dispatch_main(int fd, short event, void *arg)
 {
@@ -800,6 +838,9 @@ sockets_dispatch_main(int fd, short event, void *arg)
 		case GOTWEBD_IMSG_CTL_START:
 			sockets_launch(env);
 			break;
+		case GOTWEBD_IMSG_CTL_STOP:
+			sockets_stop();
+			break;
 		default:
 			fatalx("%s: unknown imsg type %d", __func__,
 			    imsg.hdr.type);
@@ -832,8 +873,10 @@ sockets_sighdlr(int sig, short event, void *arg)
 		break;
 	case SIGCHLD:
 		break;
-	case SIGINT:
 	case SIGTERM:
+		sockets_stop();
+		break;
+	case SIGINT:
 		sockets_shutdown();
 		break;
 	default:
@@ -1167,7 +1210,8 @@ read_fcgi_records(int fd, short events, void *arg)
 		memmove(c->buf, c->buf + record_len, c->buf_len);
 	}
 more:
-	event_add(&c->ev, NULL);
+	if (c->client_status < CLIENT_REQUEST)
+		event_add(&c->ev, NULL);
 	return;
 fail:
 	request_done(c);
@@ -1188,7 +1232,8 @@ sockets_socket_accept(int fd, short event, void *arg)
 	backoff.tv_usec = 0;
 
 	if (event & EV_TIMEOUT) {
-		event_add(&sock->ev, NULL);
+		if (!stopping)
+			event_add(&sock->ev, NULL);
 		return;
 	}
 
@@ -1202,7 +1247,8 @@ sockets_socket_accept(int fd, short event, void *arg)
 		case EINTR:
 		case EWOULDBLOCK:
 		case ECONNABORTED:
-			event_add(&sock->ev, NULL);
+			if (!stopping)
+				event_add(&sock->ev, NULL);
 			return;
 		case EMFILE:
 		case ENFILE:
@@ -1220,7 +1266,8 @@ sockets_socket_accept(int fd, short event, void *arg)
 		close(s);
 		if (c != NULL)
 			free(c);
-		event_add(&sock->ev, NULL);
+		if (!stopping)
+			event_add(&sock->ev, NULL);
 		return;
 	}
 
@@ -1229,7 +1276,8 @@ sockets_socket_accept(int fd, short event, void *arg)
 		log_warn("%s: calloc", __func__);
 		close(s);
 		cgi_inflight--;
-		event_add(&sock->ev, NULL);
+		if (!stopping)
+			event_add(&sock->ev, NULL);
 		return;
 	}
 
@@ -1239,7 +1287,8 @@ sockets_socket_accept(int fd, short event, void *arg)
 		close(s);
 		cgi_inflight--;
 		free(c);
-		event_add(&sock->ev, NULL);
+		if (!stopping)
+			event_add(&sock->ev, NULL);
 		return;
 	}
 
