@@ -14,9 +14,8 @@
  * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
  */
 
-#include "got_compat.h"
-
 #include <sys/queue.h>
+#include <sys/tree.h>
 #include <sys/stat.h>
 
 #include <errno.h>
@@ -24,9 +23,14 @@
 #include <fcntl.h>
 #include <stdio.h>
 #include <stdlib.h>
+#include <stdint.h>
 #include <string.h>
+#include <sha1.h>
+#include <sha2.h>
+#include <endian.h>
 #include <limits.h>
 #include <unistd.h>
+#include <uuid.h>
 
 #include "got_error.h"
 #include "got_object.h"
@@ -49,8 +53,10 @@
 
 struct got_fileindex {
 	struct got_fileindex_tree entries;
+	uint32_t version;
 	int nentries; /* Does not include entries marked for removal. */
-#define GOT_FILEIDX_MAX_ENTRIES INT_MAX
+#define GOT_FILEIDX_MAX_ENTRIES INT32_MAX
+	enum got_hash_algorithm algo;
 };
 
 mode_t
@@ -86,8 +92,8 @@ got_fileindex_perms_to_st(struct got_fileindex_entry *ie)
 
 const struct got_error *
 got_fileindex_entry_update(struct got_fileindex_entry *ie,
-    int wt_fd, const char *ondisk_path, uint8_t *blob_sha1,
-    uint8_t *commit_sha1, int update_timestamps)
+    int wt_fd, const char *ondisk_path, struct got_object_id *blob,
+    struct got_object_id *commit, int update_timestamps)
 {
 	struct stat sb;
 
@@ -124,14 +130,14 @@ got_fileindex_entry_update(struct got_fileindex_entry *ie,
 		}
 	}
 
-	if (blob_sha1) {
-		memmove(ie->blob_sha1, blob_sha1, SHA1_DIGEST_LENGTH);
+	if (blob) {
+		memmove(&ie->blob, blob, sizeof(ie->blob));
 		ie->flags &= ~GOT_FILEIDX_F_NO_BLOB;
 	} else
 		ie->flags |= GOT_FILEIDX_F_NO_BLOB;
 
-	if (commit_sha1) {
-		memcpy(ie->commit_sha1, commit_sha1, SHA1_DIGEST_LENGTH);
+	if (commit) {
+		memmove(&ie->commit, commit, sizeof(ie->commit));
 		ie->flags &= ~GOT_FILEIDX_F_NO_COMMIT;
 	} else
 		ie->flags |= GOT_FILEIDX_F_NO_COMMIT;
@@ -328,7 +334,7 @@ got_fileindex_for_each_entry_safe(struct got_fileindex *fileindex,
 }
 
 struct got_fileindex *
-got_fileindex_alloc(void)
+got_fileindex_alloc(enum got_hash_algorithm algo)
 {
 	struct got_fileindex *fileindex;
 
@@ -336,6 +342,8 @@ got_fileindex_alloc(void)
 	if (fileindex == NULL)
 		return NULL;
 
+	fileindex->version = GOT_FILE_INDEX_VERSION;
+	fileindex->algo = algo;
 	RB_INIT(&fileindex->entries);
 	return fileindex;
 }
@@ -421,6 +429,7 @@ write_fileindex_entry(struct got_hash *ctx, struct got_fileindex_entry *ie,
 	const struct got_error *err;
 	size_t n;
 	uint32_t stage;
+	size_t digest_len = got_hash_digest_length(ctx->algo);
 
 	err = write_fileindex_val64(ctx, ie->ctime_sec, outfile);
 	if (err)
@@ -449,14 +458,14 @@ write_fileindex_entry(struct got_hash *ctx, struct got_fileindex_entry *ie,
 	if (err)
 		return err;
 
-	got_hash_update(ctx, ie->blob_sha1, SHA1_DIGEST_LENGTH);
-	n = fwrite(ie->blob_sha1, 1, SHA1_DIGEST_LENGTH, outfile);
-	if (n != SHA1_DIGEST_LENGTH)
+	got_hash_update(ctx, ie->blob.hash, digest_len);
+	n = fwrite(ie->blob.hash, 1, digest_len, outfile);
+	if (n != digest_len)
 		return got_ferror(outfile, GOT_ERR_IO);
 
-	got_hash_update(ctx, ie->commit_sha1, SHA1_DIGEST_LENGTH);
-	n = fwrite(ie->commit_sha1, 1, SHA1_DIGEST_LENGTH, outfile);
-	if (n != SHA1_DIGEST_LENGTH)
+	got_hash_update(ctx, ie->commit.hash, digest_len);
+	n = fwrite(ie->commit.hash, 1, digest_len, outfile);
+	if (n != digest_len)
 		return got_ferror(outfile, GOT_ERR_IO);
 
 	err = write_fileindex_val32(ctx, ie->flags, outfile);
@@ -470,10 +479,10 @@ write_fileindex_entry(struct got_hash *ctx, struct got_fileindex_entry *ie,
 	stage = got_fileindex_entry_stage_get(ie);
 	if (stage == GOT_FILEIDX_STAGE_MODIFY ||
 	    stage == GOT_FILEIDX_STAGE_ADD) {
-		got_hash_update(ctx, ie->staged_blob_sha1, SHA1_DIGEST_LENGTH);
-		n = fwrite(ie->staged_blob_sha1, 1, SHA1_DIGEST_LENGTH,
+		got_hash_update(ctx, ie->staged_blob.hash, digest_len);
+		n = fwrite(ie->staged_blob.hash, 1, digest_len,
 		    outfile);
-		if (n != SHA1_DIGEST_LENGTH)
+		if (n != digest_len)
 			return got_ferror(outfile, GOT_ERR_IO);
 	}
 
@@ -487,18 +496,22 @@ got_fileindex_write(struct got_fileindex *fileindex, FILE *outfile)
 	struct got_fileindex_hdr hdr;
 	struct got_hash ctx;
 	uint8_t hash[GOT_HASH_DIGEST_MAXLEN];
-	size_t n;
+	size_t n, digest_len = got_hash_digest_length(fileindex->algo);
 	struct got_fileindex_entry *ie, *tmp;
 
-	got_hash_init(&ctx, GOT_HASH_SHA1);
+	memset(hash, 0, sizeof(hash));
+
+	got_hash_init(&ctx, fileindex->algo);
 
 	hdr.signature = htobe32(GOT_FILE_INDEX_SIGNATURE);
 	hdr.version = htobe32(GOT_FILE_INDEX_VERSION);
 	hdr.nentries = htobe32(fileindex->nentries);
+	hdr.algo = htobe32(fileindex->algo);
 
 	got_hash_update(&ctx, &hdr.signature, sizeof(hdr.signature));
 	got_hash_update(&ctx, &hdr.version, sizeof(hdr.version));
 	got_hash_update(&ctx, &hdr.nentries, sizeof(hdr.nentries));
+	got_hash_update(&ctx, &hdr.algo, sizeof(hdr.algo));
 	n = fwrite(&hdr.signature, 1, sizeof(hdr.signature), outfile);
 	if (n != sizeof(hdr.signature))
 		return got_ferror(outfile, GOT_ERR_IO);
@@ -506,6 +519,9 @@ got_fileindex_write(struct got_fileindex *fileindex, FILE *outfile)
 	if (n != sizeof(hdr.version))
 		return got_ferror(outfile, GOT_ERR_IO);
 	n = fwrite(&hdr.nentries, 1, sizeof(hdr.nentries), outfile);
+	if (n != sizeof(hdr.nentries))
+		return got_ferror(outfile, GOT_ERR_IO);
+	n = fwrite(&hdr.algo, 1, sizeof(hdr.algo), outfile);
 	if (n != sizeof(hdr.nentries))
 		return got_ferror(outfile, GOT_ERR_IO);
 
@@ -523,8 +539,8 @@ got_fileindex_write(struct got_fileindex *fileindex, FILE *outfile)
 	}
 
 	got_hash_final(&ctx, hash);
-	n = fwrite(hash, 1, SHA1_DIGEST_LENGTH, outfile);
-	if (n != SHA1_DIGEST_LENGTH)
+	n = fwrite(hash, 1, digest_len, outfile);
+	if (n != digest_len)
 		return got_ferror(outfile, GOT_ERR_IO);
 
 	if (fflush(outfile) != 0)
@@ -575,47 +591,35 @@ read_fileindex_val16(uint16_t *val, struct got_hash *ctx, FILE *infile)
 static const struct got_error *
 read_fileindex_path(char **path, struct got_hash *ctx, FILE *infile)
 {
-	const struct got_error *err = NULL;
 	const size_t chunk_size = 8;
-	size_t n, len = 0, totlen = chunk_size;
-
-	*path = malloc(totlen);
-	if (*path == NULL)
-		return got_error_from_errno("malloc");
+	char p[PATH_MAX];
+	size_t n, len = 0;
 
 	do {
-		if (len + chunk_size > totlen) {
-			char *p = reallocarray(*path, totlen + chunk_size, 1);
-			if (p == NULL) {
-				err = got_error_from_errno("reallocarray");
-				break;
-			}
-			totlen += chunk_size;
-			*path = p;
-		}
-		n = fread(*path + len, 1, chunk_size, infile);
-		if (n != chunk_size) {
-			err = got_ferror(infile, GOT_ERR_FILEIDX_BAD);
-			break;
-		}
-		got_hash_update(ctx, *path + len, chunk_size);
-		len += chunk_size;
-	} while (memchr(*path + len - chunk_size, '\0', chunk_size) == NULL);
+		if (len + chunk_size > sizeof(p))
+			return got_error(GOT_ERR_FILEIDX_BAD);
 
-	if (err) {
-		free(*path);
-		*path = NULL;
-	}
-	return err;
+		n = fread(&p[len], 1, chunk_size, infile);
+		if (n != chunk_size)
+			return got_ferror(infile, GOT_ERR_FILEIDX_BAD);
+
+		got_hash_update(ctx, &p[len], chunk_size);
+		len += chunk_size;
+	} while (memchr(&p[len - chunk_size], '\0', chunk_size) == NULL);
+
+	*path = strdup(p);
+	if (*path == NULL)
+		return got_error_from_errno("strdup");
+	return NULL;
 }
 
 static const struct got_error *
 read_fileindex_entry(struct got_fileindex_entry **iep, struct got_hash *ctx,
-    FILE *infile, uint32_t version)
+    FILE *infile, uint32_t version, enum got_hash_algorithm algo)
 {
 	const struct got_error *err;
 	struct got_fileindex_entry *ie;
-	size_t n;
+	size_t n, digest_len = got_hash_digest_length(algo);
 
 	*iep = NULL;
 
@@ -650,19 +654,21 @@ read_fileindex_entry(struct got_fileindex_entry **iep, struct got_hash *ctx,
 	if (err)
 		goto done;
 
-	n = fread(ie->blob_sha1, 1, SHA1_DIGEST_LENGTH, infile);
-	if (n != SHA1_DIGEST_LENGTH) {
+	ie->blob.algo = algo;
+	n = fread(ie->blob.hash, 1, digest_len, infile);
+	if (n != digest_len) {
 		err = got_ferror(infile, GOT_ERR_FILEIDX_BAD);
 		goto done;
 	}
-	got_hash_update(ctx, ie->blob_sha1, SHA1_DIGEST_LENGTH);
+	got_hash_update(ctx, ie->blob.hash, digest_len);
 
-	n = fread(ie->commit_sha1, 1, SHA1_DIGEST_LENGTH, infile);
-	if (n != SHA1_DIGEST_LENGTH) {
+	ie->commit.algo = algo;
+	n = fread(ie->commit.hash, 1, digest_len, infile);
+	if (n != digest_len) {
 		err = got_ferror(infile, GOT_ERR_FILEIDX_BAD);
 		goto done;
 	}
-	got_hash_update(ctx, ie->commit_sha1, SHA1_DIGEST_LENGTH);
+	got_hash_update(ctx, ie->commit.hash, digest_len);
 
 	err = read_fileindex_val32(&ie->flags, ctx, infile);
 	if (err)
@@ -676,14 +682,15 @@ read_fileindex_entry(struct got_fileindex_entry **iep, struct got_hash *ctx,
 		uint32_t stage = got_fileindex_entry_stage_get(ie);
 		if (stage == GOT_FILEIDX_STAGE_MODIFY ||
 		    stage == GOT_FILEIDX_STAGE_ADD) {
-			n = fread(ie->staged_blob_sha1, 1, SHA1_DIGEST_LENGTH,
+			ie->staged_blob.algo = algo;
+			n = fread(ie->staged_blob.hash, 1, digest_len,
 			    infile);
-			if (n != SHA1_DIGEST_LENGTH) {
+			if (n != digest_len) {
 				err = got_ferror(infile, GOT_ERR_FILEIDX_BAD);
 				goto done;
 			}
-			got_hash_update(ctx, ie->staged_blob_sha1,
-			    SHA1_DIGEST_LENGTH);
+			got_hash_update(ctx, ie->staged_blob.hash,
+			    digest_len);
 		}
 	} else {
 		/* GOT_FILE_INDEX_VERSION 1 does not support staging. */
@@ -699,19 +706,20 @@ done:
 }
 
 const struct got_error *
-got_fileindex_read(struct got_fileindex *fileindex, FILE *infile)
+got_fileindex_read(struct got_fileindex *fileindex, FILE *infile,
+    enum got_hash_algorithm repo_algo)
 {
 	const struct got_error *err = NULL;
 	struct got_fileindex_hdr hdr;
 	struct got_hash ctx;
 	struct got_fileindex_entry *ie;
-	uint8_t sha1_expected[SHA1_DIGEST_LENGTH];
-	uint8_t sha1[SHA1_DIGEST_LENGTH];
-	size_t n;
+	enum got_hash_algorithm algo = repo_algo;
+	uint8_t hash_expected[GOT_HASH_DIGEST_MAXLEN];
+	uint8_t hash[GOT_HASH_DIGEST_MAXLEN];
+	size_t n, digest_len;
+	uint32_t version;
 	int i;
 
-	got_hash_init(&ctx, GOT_HASH_SHA1);
-
 	n = fread(&hdr.signature, 1, sizeof(hdr.signature), infile);
 	if (n != sizeof(hdr.signature)) {
 		if (n == 0) /* EOF */
@@ -730,10 +738,36 @@ got_fileindex_read(struct got_fileindex *fileindex, FILE *infile)
 			return NULL;
 		return got_ferror(infile, GOT_ERR_FILEIDX_BAD);
 	}
+	version = be32toh(hdr.version);
 
+	if (version >= 3) {
+		n = fread(&hdr.algo, 1, sizeof(hdr.algo), infile);
+		if (n != sizeof(hdr.algo)) {
+			if (n == 0) /* EOF */
+				return NULL;
+			return got_ferror(infile, GOT_ERR_FILEIDX_BAD);
+		}
+		algo = be32toh(hdr.algo);
+		if (algo != repo_algo) {
+			const char *fmt = "unknown";
+
+			if (repo_algo == GOT_HASH_SHA1)
+				fmt = "sha1";
+			else if (repo_algo == GOT_HASH_SHA256)
+				fmt = "sha256";
+
+			return got_error_path(fmt, GOT_ERR_OBJECT_FORMAT);
+		}
+	}
+
+	digest_len = got_hash_digest_length(algo);
+
+	got_hash_init(&ctx, algo);
 	got_hash_update(&ctx, &hdr.signature, sizeof(hdr.signature));
 	got_hash_update(&ctx, &hdr.version, sizeof(hdr.version));
 	got_hash_update(&ctx, &hdr.nentries, sizeof(hdr.nentries));
+	if (version >= 3)
+		got_hash_update(&ctx, &hdr.algo, sizeof(hdr.algo));
 
 	hdr.signature = be32toh(hdr.signature);
 	hdr.version = be32toh(hdr.version);
@@ -744,25 +778,36 @@ got_fileindex_read(struct got_fileindex *fileindex, FILE *infile)
 	if (hdr.version > GOT_FILE_INDEX_VERSION)
 		return got_error(GOT_ERR_FILEIDX_VER);
 
+	fileindex->version = version;
+	fileindex->algo = algo;
 	for (i = 0; i < hdr.nentries; i++) {
-		err = read_fileindex_entry(&ie, &ctx, infile, hdr.version);
+		err = read_fileindex_entry(&ie, &ctx, infile,
+		    hdr.version, algo);
 		if (err)
 			return err;
 		err = add_entry(fileindex, ie);
-		if (err)
+		if (err) {
+			got_fileindex_entry_free(ie);
 			return err;
+		}
 	}
 
-	n = fread(sha1_expected, 1, sizeof(sha1_expected), infile);
-	if (n != sizeof(sha1_expected))
+	n = fread(hash_expected, 1, digest_len, infile);
+	if (n != digest_len)
 		return got_ferror(infile, GOT_ERR_FILEIDX_BAD);
-	got_hash_final(&ctx, sha1);
-	if (memcmp(sha1, sha1_expected, SHA1_DIGEST_LENGTH) != 0)
+	got_hash_final(&ctx, hash);
+	if (got_hash_cmp(algo, hash, hash_expected) != 0)
 		return got_error(GOT_ERR_FILEIDX_CSUM);
 
 	return NULL;
 }
 
+uint32_t
+got_fileindex_get_version(struct got_fileindex *fileindex)
+{
+	return fileindex->version;
+}
+
 static struct got_fileindex_entry *
 walk_fileindex(struct got_fileindex *fileindex, struct got_fileindex_entry *ie)
 {
@@ -931,39 +976,49 @@ diff_fileindex_dir(struct got_fileindex *, struct got_fileindex_entry **,
     struct got_pathlist_head *, int, const char *, const char *,
     struct got_repository *, struct got_fileindex_diff_dir_cb *, void *);
 
+static struct dirent *
+copy_dirent(const struct dirent *de)
+{
+	size_t amt = de->d_reclen;
+	struct dirent *copy;
+
+	copy = malloc(amt);
+	if (copy != NULL) {
+		memcpy(copy, de, amt);
+	}
+	return copy;
+}
+
 static const struct got_error *
 read_dirlist(struct got_pathlist_head *dirlist, DIR *dir, const char *path)
 {
 	const struct got_error *err = NULL;
 	struct got_pathlist_entry *new = NULL;
-	struct dirent *dep = NULL;
 	struct dirent *de = NULL;
 
 	for (;;) {
-		de = malloc(sizeof(struct dirent) + NAME_MAX + 1);
-		if (de == NULL) {
-			err = got_error_from_errno("malloc");
-			break;
-		}
-
-		if (readdir_r(dir, de, &dep) != 0) {
-			err = got_error_from_errno("readdir_r");
-			free(de);
-			break;
-		}
-		if (dep == NULL) {
-			free(de);
+		errno = 0;
+		if ((de = readdir(dir)) == NULL) {
+			if (errno != 0) {
+				err = got_error_from_errno("readdir");
+			}
 			break;
 		}
 
 		if (strcmp(de->d_name, ".") == 0 ||
 		    strcmp(de->d_name, "..") == 0 ||
 		    (path[0] == '\0' &&
-		    strcmp(de->d_name, GOT_WORKTREE_GOT_DIR) == 0)) {
-			free(de);
+		    strcmp(de->d_name, GOT_WORKTREE_GOT_DIR) == 0) ||
+		    (path[0] == '\0' &&
+		    strcmp(de->d_name, GOT_WORKTREE_CVG_DIR) == 0)) {
 			continue;
 		}
 
+		de = copy_dirent(de);
+		if (de == NULL) {
+			err = got_error_from_errno("malloc");
+			break;
+		}
 		err = got_pathlist_insert(&new, dirlist, de->d_name, de);
 		if (err) {
 			free(de);
@@ -1005,9 +1060,10 @@ have_tracked_file_in_dir(struct got_fileindex *fileindex, const char *path)
 
 static const struct got_error *
 walk_dir(struct got_pathlist_entry **next, struct got_fileindex *fileindex,
-    struct got_fileindex_entry **ie, struct got_pathlist_entry *dle, int fd,
-    const char *path, const char *rootpath, struct got_repository *repo,
-    int ignore, struct got_fileindex_diff_dir_cb *cb, void *cb_arg)
+    struct got_fileindex_entry **ie, struct got_pathlist_entry *dle,
+    struct got_pathlist_head *dlh, int fd, const char *path,
+    const char *rootpath, struct got_repository *repo, int ignore,
+    struct got_fileindex_diff_dir_cb *cb, void *cb_arg)
 {
 	const struct got_error *err = NULL;
 	struct dirent *de = dle->data;
@@ -1026,7 +1082,7 @@ walk_dir(struct got_pathlist_entry **next, struct got_fileindex *fileindex,
 		char *subdirpath;
 		struct got_pathlist_head subdirlist;
 
-		TAILQ_INIT(&subdirlist);
+		RB_INIT(&subdirlist);
 
 		if (asprintf(&subpath, "%s%s%s", path,
 		    path[0] == '\0' ? "" : "/", de->d_name) == -1)
@@ -1041,7 +1097,7 @@ walk_dir(struct got_pathlist_entry **next, struct got_fileindex *fileindex,
 		    O_RDONLY | O_NOFOLLOW | O_DIRECTORY | O_CLOEXEC);
 		if (subdirfd == -1) {
 			if (errno == EACCES) {
-				*next = TAILQ_NEXT(dle, entry);
+				*next = RB_NEXT(got_pathlist_head, dlh, dle);
 				return NULL;
 			}
 			err = got_error_from_errno2("openat", subdirpath);
@@ -1051,8 +1107,13 @@ walk_dir(struct got_pathlist_entry **next, struct got_fileindex *fileindex,
 		}
 
 		subdir = fdopendir(subdirfd);
-		if (subdir == NULL)
-			return got_error_from_errno2("fdopendir", path);
+		if (subdir == NULL) {
+			err = got_error_from_errno2("fdopendir", path);
+			close(subdirfd);
+			free(subpath);
+			free(subdirpath);
+			return err;
+		}
 		subdirfd = -1;
 		err = read_dirlist(&subdirlist, subdir, subdirpath);
 		if (err) {
@@ -1072,7 +1133,7 @@ walk_dir(struct got_pathlist_entry **next, struct got_fileindex *fileindex,
 			return err;
 	}
 
-	*next = TAILQ_NEXT(dle, entry);
+	*next = RB_NEXT(got_pathlist_head, dlh, dle);
 	return NULL;
 }
 
@@ -1117,7 +1178,7 @@ diff_fileindex_dir(struct got_fileindex *fileindex,
 			return err;
 	}
 
-	dle = TAILQ_FIRST(dirlist);
+	dle = RB_MIN(got_pathlist_head, dirlist);
 	while ((*ie && got_path_is_child((*ie)->path, path, path_len)) || dle) {
 		if (dle && *ie) {
 			char *de_path;
@@ -1133,7 +1194,7 @@ diff_fileindex_dir(struct got_fileindex *fileindex,
 			}
 			cmp = got_path_cmp((*ie)->path, de_path,
 			    got_fileindex_entry_path_len(*ie),
-			    strlen(path) + 1 + strlen(de->d_name));
+			    strlen(path) + 1 + de->d_namlen);
 			free(de_path);
 			if (cmp == 0) {
 				err = cb->diff_old_new(cb_arg, *ie, de, path,
@@ -1141,9 +1202,10 @@ diff_fileindex_dir(struct got_fileindex *fileindex,
 				if (err)
 					break;
 				*ie = walk_fileindex(fileindex, *ie);
-				err = walk_dir(&dle, fileindex, ie, dle, dirfd,
-				    path, rootpath, repo, 0, cb, cb_arg);
-			} else if (cmp < 0 ) {
+				err = walk_dir(&dle, fileindex, ie, dle,
+				    dirlist, dirfd, path, rootpath, repo, 0,
+				    cb, cb_arg);
+			} else if (cmp < 0) {
 				err = cb->diff_old(cb_arg, *ie, path);
 				if (err)
 					break;
@@ -1153,8 +1215,9 @@ diff_fileindex_dir(struct got_fileindex *fileindex,
 				    dirfd);
 				if (err)
 					break;
-				err = walk_dir(&dle, fileindex, ie, dle, dirfd,
-				    path, rootpath, repo, ignore, cb, cb_arg);
+				err = walk_dir(&dle, fileindex, ie, dle,
+				    dirlist, dirfd, path, rootpath, repo,
+				    ignore, cb, cb_arg);
 			}
 			if (err)
 				break;
@@ -1171,8 +1234,8 @@ diff_fileindex_dir(struct got_fileindex *fileindex,
 			err = cb->diff_new(&ignore, cb_arg, de, path, dirfd);
 			if (err)
 				break;
-			err = walk_dir(&dle, fileindex, ie, dle, dirfd, path,
-			    rootpath, repo, ignore, cb, cb_arg);
+			err = walk_dir(&dle, fileindex, ie, dle, dirlist,
+			    dirfd, path, rootpath, repo, ignore, cb, cb_arg);
 			if (err)
 				break;
 		}
@@ -1192,7 +1255,7 @@ got_fileindex_diff_dir(struct got_fileindex *fileindex, int fd,
 	int fd2;
 	DIR *dir;
 
-	TAILQ_INIT(&dirlist);
+	RB_INIT(&dirlist);
 
 	/*
 	 * Duplicate the file descriptor so we can call closedir() below
@@ -1234,27 +1297,21 @@ struct got_object_id *
 got_fileindex_entry_get_staged_blob_id(struct got_object_id *id,
     struct got_fileindex_entry *ie)
 {
-	memset(id, 0, sizeof(*id));
-	memcpy(id->sha1, ie->staged_blob_sha1, sizeof(ie->staged_blob_sha1));
-	return id;
+	return memcpy(id, &ie->staged_blob, sizeof(*id));
 }
 
 struct got_object_id *
 got_fileindex_entry_get_blob_id(struct got_object_id *id,
     struct got_fileindex_entry *ie)
 {
-	memset(id, 0, sizeof(*id));
-	memcpy(id->sha1, ie->blob_sha1, sizeof(ie->blob_sha1));
-	return id;
+	return memcpy(id, &ie->blob, sizeof(*id));
 }
 
 struct got_object_id *
 got_fileindex_entry_get_commit_id(struct got_object_id *id,
     struct got_fileindex_entry *ie)
 {
-	memset(id, 0, sizeof(*id));
-	memcpy(id->sha1, ie->commit_sha1, sizeof(ie->commit_sha1));
-	return id;
+	return memcpy(id, &ie->commit, sizeof(*id));
 }
 
 RB_GENERATE(got_fileindex_tree, got_fileindex_entry, entry, got_fileindex_cmp);
