@@ -17,20 +17,23 @@
 #include <sys/mman.h>
 #include <sys/queue.h>
 #include <sys/types.h>
+#include <sys/tree.h>
 #include <sys/stat.h>
 #include <sys/socket.h>
 #include <sys/uio.h>
 
 #include <errno.h>
 #include <imsg.h>
+#include <stddef.h>
 #include <stdio.h>
 #include <stdint.h>
 #include <stdlib.h>
 #include <string.h>
+#include <sha1.h>
+#include <sha2.h>
 #include <limits.h>
 #include <unistd.h>
 
-#include "got_compat.h"
 #include "got_error.h"
 #include "got_object.h"
 #include "got_repository.h"
@@ -39,6 +42,7 @@
 
 #include "got_lib_delta.h"
 #include "got_lib_object.h"
+#include "got_lib_hash.h"
 #include "got_lib_privsep.h"
 #include "got_lib_object_cache.h"
 #include "got_lib_pack.h"
@@ -114,8 +118,8 @@ request_packed_object_raw(uint8_t **outbuf, off_t *size, size_t *hdrlen,
 	struct imsgbuf *ibuf = pack->privsep_child->ibuf;
 	int outfd_child;
 
-	 err = pack_child_send_tempfiles(ibuf, pack);
-	 if (err)
+	err = pack_child_send_tempfiles(ibuf, pack);
+	if (err)
 		return err;
 
 	outfd_child = dup(outfd);
@@ -356,6 +360,8 @@ start_child(struct got_repository *repo, int type)
 	pid = fork();
 	if (pid == -1) {
 		err = got_error_from_errno("fork");
+		close(imsg_fds[0]);
+		close(imsg_fds[1]);
 		free(ibuf);
 		return err;
 	}
@@ -366,13 +372,21 @@ start_child(struct got_repository *repo, int type)
 
 	if (close(imsg_fds[1]) == -1) {
 		err = got_error_from_errno("close");
+		close(imsg_fds[0]);
 		free(ibuf);
 		return err;
 	}
 
 	repo->privsep_children[type].imsg_fd = imsg_fds[0];
 	repo->privsep_children[type].pid = pid;
-	imsg_init(ibuf, imsg_fds[0]);
+	if (imsgbuf_init(ibuf, imsg_fds[0]) == -1) {
+		err = got_error_from_errno("imsgbuf_init");
+		close(imsg_fds[0]);
+		free(ibuf);
+		return err;
+	}
+	imsgbuf_allow_fdpass(ibuf);
+
 	repo->privsep_children[type].ibuf = ibuf;
 
 	return NULL;
@@ -435,11 +449,8 @@ got_object_open(struct got_object **obj, struct got_repository *repo,
 	}
 
 	err = got_object_open_loose_fd(&fd, id, repo);
-	if (err) {
-		if (err->code == GOT_ERR_ERRNO && errno == ENOENT)
-			err = got_error_no_obj(id);
+	if (err)
 		return err;
-	}
 
 	err = got_object_read_header_privsep(obj, id, repo, fd);
 	if (err)
@@ -814,8 +825,8 @@ request_packed_blob(uint8_t **outbuf, size_t *size, size_t *hdrlen, int outfd,
 	struct imsgbuf *ibuf = pack->privsep_child->ibuf;
 	int outfd_child;
 
-	 err = pack_child_send_tempfiles(ibuf, pack);
-	 if (err)
+	err = pack_child_send_tempfiles(ibuf, pack);
+	if (err)
 		return err;
 
 	outfd_child = dup(outfd);
@@ -837,7 +848,7 @@ request_packed_blob(uint8_t **outbuf, size_t *size, size_t *hdrlen, int outfd,
 	if (err)
 		return err;
 
-	if (lseek(outfd, SEEK_SET, 0) == -1)
+	if (lseek(outfd, 0L, SEEK_SET) == -1)
 		err = got_error_from_errno("lseek");
 
 	return err;
@@ -883,7 +894,7 @@ request_blob(uint8_t **outbuf, size_t *size, size_t *hdrlen, int outfd,
 	if (err)
 		return err;
 
-	if (lseek(outfd, SEEK_SET, 0) == -1)
+	if (lseek(outfd, 0L, SEEK_SET) == -1)
 		return got_error_from_errno("lseek");
 
 	return err;
@@ -932,14 +943,9 @@ open_blob(struct got_blob_object **blob, struct got_repository *repo,
 		goto done;
 	}
 
-	if (ftruncate(outfd, 0L) == -1) {
-		err = got_error_from_errno("ftruncate");
+	err = got_opentemp_truncatefd(outfd);
+	if (err)
 		goto done;
-	}
-	if (lseek(outfd, SEEK_SET, 0) == -1) {
-		err = got_error_from_errno("lseek");
-		goto done;
-	}
 
 	err = got_repo_search_packidx(&packidx, &idx, repo, id);
 	if (err == NULL) {
@@ -976,7 +982,7 @@ open_blob(struct got_blob_object **blob, struct got_repository *repo,
 		goto done;
 	}
 
-	if (outbuf) {
+	if (outbuf && size > 0) {
 		(*blob)->f = fmemopen(outbuf, size, "rb");
 		if ((*blob)->f == NULL) {
 			err = got_error_from_errno("fmemopen");
@@ -1212,7 +1218,7 @@ got_traverse_packed_commits(struct got_object_id_queue *traversed_commits,
 	struct got_packidx *packidx = NULL;
 	char *path_packfile = NULL;
 	struct got_commit_object *changed_commit = NULL;
-	struct got_object_id *changed_commit_id = NULL;
+	struct got_object_qid *changed_commit_qid = NULL;
 	int idx;
 
 	err = got_repo_search_packidx(&packidx, &idx, repo, commit_id);
@@ -1246,7 +1252,7 @@ got_traverse_packed_commits(struct got_object_id_queue *traversed_commits,
 		goto done;
 
 	err = got_privsep_recv_traversed_commits(&changed_commit,
-	    &changed_commit_id, traversed_commits, pack->privsep_child->ibuf);
+	    traversed_commits, pack->privsep_child->ibuf);
 	if (err)
 		goto done;
 
@@ -1256,13 +1262,13 @@ got_traverse_packed_commits(struct got_object_id_queue *traversed_commits,
 		 * This commit might be opened again soon.
 		 */
 		changed_commit->refcnt++;
-		err = got_repo_cache_commit(repo, changed_commit_id,
+		changed_commit_qid = STAILQ_LAST(traversed_commits, got_object_qid, entry);
+		err = got_repo_cache_commit(repo, &changed_commit_qid->id,
 		    changed_commit);
 		got_object_commit_close(changed_commit);
 	}
 done:
 	free(path_packfile);
-	free(changed_commit_id);
 	return err;
 }
 
