@@ -16,12 +16,15 @@
 
 #include <sys/types.h>
 #include <sys/queue.h>
+#include <sys/tree.h>
 #include <sys/uio.h>
 
 #include <errno.h>
 #include <event.h>
 #include <poll.h>
 #include <limits.h>
+#include <sha1.h>
+#include <sha2.h>
 #include <stdio.h>
 #include <stdint.h>
 #include <stdlib.h>
@@ -48,6 +51,12 @@
 #define nitems(_a)	(sizeof((_a)) / sizeof((_a)[0]))
 #endif
 
+/*
+ * Number of seconds until we give up trying to read more
+ * data from the client connection.
+ */
+static const int timeout = 60;
+
 static const struct got_capability read_capabilities[] = {
 	{ GOT_CAPA_AGENT, "got/" GOT_VERSION_STR },
 	{ GOT_CAPA_OFS_DELTA, NULL },
@@ -62,103 +71,6 @@ static const struct got_capability write_capabilities[] = {
 	{ GOT_CAPA_DELETE_REFS, NULL },
 };
 
-const struct got_error *
-got_serve_parse_command(char **command, char **repo_path, const char *gitcmd)
-{
-	const struct got_error *err = NULL;
-	size_t len, cmdlen, pathlen;
-	char *path0 = NULL, *path, *abspath = NULL, *canonpath = NULL;
-	const char *relpath;
-
-	*command = NULL;
-	*repo_path = NULL;
-
-	len = strlen(gitcmd);
-
-	if (len >= strlen(GOT_SERVE_CMD_SEND) &&
-	    strncmp(gitcmd, GOT_SERVE_CMD_SEND,
-	    strlen(GOT_SERVE_CMD_SEND)) == 0)
-		cmdlen = strlen(GOT_SERVE_CMD_SEND);
-	else if (len >= strlen(GOT_SERVE_CMD_FETCH) &&
-	    strncmp(gitcmd, GOT_SERVE_CMD_FETCH,
-	    strlen(GOT_SERVE_CMD_FETCH)) == 0)
-		cmdlen = strlen(GOT_SERVE_CMD_FETCH);
-	else
-		return got_error(GOT_ERR_BAD_PACKET);
-
-	if (len <= cmdlen + 1 || gitcmd[cmdlen] != ' ')
-		return got_error(GOT_ERR_BAD_PACKET);
-
-	if (memchr(&gitcmd[cmdlen + 1], '\0', len - cmdlen) == NULL)
-		return got_error(GOT_ERR_BAD_PATH);
-
-	/* Forbid linefeeds in paths, like Git does. */
-	if (memchr(&gitcmd[cmdlen + 1], '\n', len - cmdlen) != NULL)
-		return got_error(GOT_ERR_BAD_PATH);
-
-	path0 = strdup(&gitcmd[cmdlen + 1]);
-	if (path0 == NULL)
-		return got_error_from_errno("strdup");
-	path = path0;
-	pathlen = strlen(path);
-
-	/*
-	 * Git clients send a shell command.
-	 * Trim spaces and quotes around the path.
-	 */
-	while (path[0] == '\'' || path[0] == '\"' || path[0] == ' ') {
-		path++;
-		pathlen--;
-	}
-	while (pathlen > 0 &&
-	    (path[pathlen - 1] == '\'' || path[pathlen - 1] == '\"' ||
-	    path[pathlen - 1] == ' ')) {
-		path[pathlen - 1] = '\0';
-		pathlen--;
-	}
-
-	/* Deny an empty repository path. */
-	if (path[0] == '\0' || got_path_is_root_dir(path)) {
-		err = got_error(GOT_ERR_NOT_GIT_REPO);
-		goto done;
-	}
-
-	if (asprintf(&abspath, "/%s", path) == -1) {
-		err = got_error_from_errno("asprintf");
-		goto done;
-	}
-	pathlen = strlen(abspath);
-	canonpath = malloc(pathlen);
-	if (canonpath == NULL) {
-		err = got_error_from_errno("malloc");
-		goto done;
-	}
-	err = got_canonpath(abspath, canonpath, pathlen);
-	if (err)
-		goto done;
-
-	relpath = canonpath;
-	while (relpath[0] == '/')
-		relpath++;
-	*repo_path = strdup(relpath);
-	if (*repo_path == NULL) {
-		err = got_error_from_errno("strdup");
-		goto done;
-	}
-	*command = strndup(gitcmd, cmdlen);
-	if (*command == NULL)
-		err = got_error_from_errno("strndup");
-done:
-	free(path0);
-	free(abspath);
-	free(canonpath);
-	if (err) {
-		free(*repo_path);
-		*repo_path = NULL;
-	}
-	return err;
-}
-
 static const struct got_error *
 append_read_capabilities(size_t *capalen, size_t len, const char *symrefstr,
     uint8_t *buf, size_t bufsize)
@@ -289,8 +201,20 @@ announce_refs(int outfd, struct imsgbuf *ibuf, int client_is_reading,
 		return got_error_from_errno("imsg_compose LIST_REFS");
 
 	err = gotd_imsg_flush(ibuf);
-	if (err)
+	if (err) {
+		if (err->code == GOT_ERR_EOF) {
+			const struct got_error *err2;
+
+			/* Connection closed early. Try to find out why. */
+			err2 = gotd_imsg_poll_recv(&imsg, ibuf, 0);
+			if (err2)
+				return err;
+			if (imsg.hdr.type == GOTD_IMSG_ERROR)
+				err = gotd_imsg_recv_error(NULL, &imsg);
+			imsg_free(&imsg);
+		}
 		return err;
+	}
 
 	while (!have_nrefs || nrefs > 0) {
 		err = gotd_imsg_poll_recv(&imsg, ibuf, 0);
@@ -503,7 +427,6 @@ send_capability(struct got_capability *capa, struct imsgbuf *ibuf)
 			return got_error_from_errno("imsg_add CAPABILITY");
 	}
 
-	wbuf->fd = -1;
 	imsg_close(ibuf, wbuf);
 
 	return NULL;
@@ -761,6 +684,7 @@ recv_done(int *packfd, int outfd, struct imsgbuf *ibuf, int chattygot)
 {
 	const struct got_error *err;
 	struct imsg imsg;
+	int fd;
 
 	*packfd = -1;
 
@@ -781,10 +705,17 @@ recv_done(int *packfd, int outfd, struct imsgbuf *ibuf, int chattygot)
 			err = gotd_imsg_recv_error(NULL, &imsg);
 			break;
 		case GOTD_IMSG_PACKFILE_PIPE:
-			if (imsg.fd != -1)
-				*packfd = imsg.fd;
-			else
+			fd = imsg_get_fd(&imsg);
+			if (fd == -1) {
 				err = got_error(GOT_ERR_PRIVSEP_NO_FD);
+				break;
+			}
+			*packfd = fd;
+			if (imsg_compose(ibuf, GOTD_IMSG_PACKFILE_READY,
+			    0, 0, -1, NULL, 0) == -1)
+				err = got_error_from_errno("imsg_compose "
+				    "PACKFILE_READY");
+			err = gotd_imsg_flush(ibuf);
 			break;
 		default:
 			err = got_error(GOT_ERR_PRIVSEP_MSG);
@@ -868,8 +799,7 @@ serve_read(int infd, int outfd, int gotd_sock, const char *repo_path,
 	enum protostate {
 		STATE_EXPECT_WANT,
 		STATE_EXPECT_MORE_WANT,
-		STATE_EXPECT_HAVE,
-		STATE_EXPECT_DONE,
+		STATE_EXPECT_HAVE_OR_DONE,
 		STATE_DONE,
 	};
 	enum protostate curstate = STATE_EXPECT_WANT;
@@ -877,7 +807,9 @@ serve_read(int infd, int outfd, int gotd_sock, const char *repo_path,
 	int packfd = -1;
 	size_t pack_chunksize;
 
-	imsg_init(&ibuf, gotd_sock);
+	if (imsgbuf_init(&ibuf, gotd_sock) == -1)
+		return got_error_from_errno("imsgbuf_init");
+	imsgbuf_allow_fdpass(&ibuf);
 
 	err = announce_refs(outfd, &ibuf, 1, repo_path, chattygot);
 	if (err)
@@ -886,14 +818,14 @@ serve_read(int infd, int outfd, int gotd_sock, const char *repo_path,
 	while (curstate != STATE_DONE) {
 		int n;
 		buf[0] = '\0';
-		err = got_pkt_readpkt(&n, infd, buf, sizeof(buf), chattygot);
+		err = got_pkt_readpkt(&n, infd, buf, sizeof(buf), chattygot,
+		    timeout);
 		if (err)
 			goto done;
 		if (n == 0) {
 			if (curstate != STATE_EXPECT_WANT &&
 			    curstate != STATE_EXPECT_MORE_WANT &&
-			    curstate != STATE_EXPECT_HAVE &&
-			    curstate != STATE_EXPECT_DONE) {
+			    curstate != STATE_EXPECT_HAVE_OR_DONE) {
 				err = got_error_msg(GOT_ERR_BAD_PACKET,
 				    "unexpected flush packet received");
 				goto done;
@@ -922,20 +854,19 @@ serve_read(int infd, int outfd, int gotd_sock, const char *repo_path,
 
 			if (curstate == STATE_EXPECT_WANT ||
 			    curstate == STATE_EXPECT_MORE_WANT ||
-			    curstate == STATE_EXPECT_HAVE) {
+			    curstate == STATE_EXPECT_HAVE_OR_DONE) {
 				err = forward_flushpkt(&ibuf);
 				if (err)
 					goto done;
 			}
-			if (curstate == STATE_EXPECT_HAVE && !have_ack) {
+			if (curstate == STATE_EXPECT_HAVE_OR_DONE &&
+			    !have_ack) {
 				err = send_nak(outfd, chattygot);
 				if (err)
 					goto done;
 			}
 			if (curstate == STATE_EXPECT_MORE_WANT)
-				curstate = STATE_EXPECT_HAVE;
-			else
-				curstate = STATE_EXPECT_DONE;
+				curstate = STATE_EXPECT_HAVE_OR_DONE;
 		} else if (n >= 5 && strncmp(buf, "want ", 5) == 0) {
 			if (curstate != STATE_EXPECT_WANT &&
 			    curstate != STATE_EXPECT_MORE_WANT) {
@@ -950,22 +881,18 @@ serve_read(int infd, int outfd, int gotd_sock, const char *repo_path,
 			if (curstate == STATE_EXPECT_WANT)
 				curstate = STATE_EXPECT_MORE_WANT;
 		} else if (n >= 5 && strncmp(buf, "have ", 5) == 0) {
-			if (curstate != STATE_EXPECT_HAVE &&
-			    curstate != STATE_EXPECT_DONE) {
+			if (curstate != STATE_EXPECT_HAVE_OR_DONE) {
 				err = got_error_msg(GOT_ERR_BAD_PACKET,
 				    "unexpected 'have' packet");
 				goto done;
 			}
-			if (curstate == STATE_EXPECT_HAVE) {
-				err = recv_have(&have_ack, outfd, &ibuf,
-				    buf, n, chattygot);
-				if (err)
-					goto done;
-				seen_have = 1;
-			}
+			err = recv_have(&have_ack, outfd, &ibuf,
+			    buf, n, chattygot);
+			if (err)
+				goto done;
+			seen_have = 1;
 		} else if (n == 5 && strncmp(buf, "done\n", 5) == 0) {
-			if (curstate != STATE_EXPECT_HAVE &&
-			    curstate != STATE_EXPECT_DONE) {
+			if (curstate != STATE_EXPECT_HAVE_OR_DONE) {
 				err = got_error_msg(GOT_ERR_BAD_PACKET,
 				    "unexpected 'done' packet");
 				goto done;
@@ -1023,7 +950,7 @@ serve_read(int infd, int outfd, int gotd_sock, const char *repo_path,
 		}
 	}
 done:
-	imsg_clear(&ibuf);
+	imsgbuf_clear(&ibuf);
 	if (packfd != -1 && close(packfd) == -1 && err == NULL)
 		err = got_error_from_errno("close");
 	if (err)
@@ -1119,7 +1046,6 @@ recv_ref_update(int *report_status, int outfd, struct imsgbuf *ibuf,
 		return got_error_from_errno("imsg_add REF_UPDATE");
 	if (imsg_add(wbuf, refname, iref.name_len) == -1)
 		return got_error_from_errno("imsg_add REF_UPDATE");
-	wbuf->fd = -1;
 	imsg_close(ibuf, wbuf);
 
 	err = gotd_imsg_flush(ibuf);
@@ -1155,7 +1081,7 @@ done:
 }
 
 static const struct got_error *
-recv_packfile(struct imsg *imsg, int infd)
+recv_packfile(struct imsg *imsg, struct imsgbuf *ibuf, int infd)
 {
 	const struct got_error *err = NULL;
 	size_t datalen;
@@ -1167,10 +1093,18 @@ recv_packfile(struct imsg *imsg, int infd)
 	if (datalen != 0)
 		return got_error(GOT_ERR_PRIVSEP_MSG);
 
-	if (imsg->fd == -1)
+	packfd = imsg_get_fd(imsg);
+	if (packfd == -1)
 		return got_error(GOT_ERR_PRIVSEP_NO_FD);
 
-	packfd = imsg->fd;
+	if (imsg_compose(ibuf, GOTD_IMSG_PACKFILE_READY,
+	    0, 0, -1, NULL, 0) == -1)
+		return got_error_from_errno("imsg_compose PACKFILE_READY");
+
+	err = gotd_imsg_flush(ibuf);
+	if (err)
+		return err;
+
 	while (!pack_done) {
 		ssize_t r = 0;
 
@@ -1198,12 +1132,14 @@ recv_packfile(struct imsg *imsg, int infd)
 
 		if (r == 0) {
 			/* Detect gotd(8) closing the pack pipe when done. */
-			err = got_poll_fd(packfd, POLLOUT, 1);
+			err = got_poll_fd(packfd, 0, 1);
 			if (err) {
-				if (err->code != GOT_ERR_EOF)
+				if (err->code != GOT_ERR_TIMEOUT &&
+				    err->code != GOT_ERR_EOF)
 					break;
+				if (err->code == GOT_ERR_EOF)
+					pack_done = 1;
 				err = NULL;
-				pack_done = 1;
 			}
 		} else {
 			/* Write pack data and/or detect pipe being closed. */
@@ -1351,7 +1287,10 @@ serve_write(int infd, int outfd, int gotd_sock, const char *repo_path,
 	struct imsg imsg;
 	int report_status = 0;
 
-	imsg_init(&ibuf, gotd_sock);
+	if (imsgbuf_init(&ibuf, gotd_sock) == -1)
+		return got_error_from_errno("imsgbuf_init");
+	imsgbuf_allow_fdpass(&ibuf);
+
 	memset(&imsg, 0, sizeof(imsg));
 
 	err = announce_refs(outfd, &ibuf, 0, repo_path, chattygot);
@@ -1361,11 +1300,15 @@ serve_write(int infd, int outfd, int gotd_sock, const char *repo_path,
 	while (curstate != STATE_EXPECT_PACKFILE) {
 		int n;
 		buf[0] = '\0';
-		err = got_pkt_readpkt(&n, infd, buf, sizeof(buf), chattygot);
+		err = got_pkt_readpkt(&n, infd, buf, sizeof(buf), chattygot,
+		    timeout);
 		if (err)
 			goto done;
 		if (n == 0) {
-			if (curstate != STATE_EXPECT_MORE_REF_UPDATES) {
+			if (curstate == STATE_EXPECT_REF_UPDATE) {
+				/* The client will not send us anything. */
+				goto done;
+			} else if (curstate != STATE_EXPECT_MORE_REF_UPDATES) {
 				err = got_error_msg(GOT_ERR_BAD_PACKET,
 				    "unexpected flush packet received");
 				goto done;
@@ -1405,8 +1348,8 @@ serve_write(int infd, int outfd, int gotd_sock, const char *repo_path,
 		case GOTD_IMSG_ERROR:
 			err = gotd_imsg_recv_error(NULL, &imsg);
 			goto done;
-		case GOTD_IMSG_RECV_PACKFILE:
-			err = recv_packfile(&imsg, infd);
+		case GOTD_IMSG_PACKFILE_PIPE:
+			err = recv_packfile(&imsg, &ibuf, infd);
 			if (err) {
 				if (err->code != GOT_ERR_EOF)
 					goto done;
@@ -1465,7 +1408,7 @@ serve_write(int infd, int outfd, int gotd_sock, const char *repo_path,
 		imsg_free(&imsg);
 	}
 done:
-	imsg_clear(&ibuf);
+	imsgbuf_clear(&ibuf);
 	if (err)
 		echo_error(err, outfd, chattygot);
 	return err;
@@ -1477,9 +1420,9 @@ got_serve(int infd, int outfd, const char *command, const char *repo_path,
 {
 	const struct got_error *err = NULL;
 
-	if (strcmp(command, GOT_SERVE_CMD_FETCH) == 0)
+	if (strcmp(command, GOT_DIAL_CMD_FETCH) == 0)
 		err = serve_read(infd, outfd, gotd_sock, repo_path, chattygot);
-	else if (strcmp(command, GOT_SERVE_CMD_SEND) == 0)
+	else if (strcmp(command, GOT_DIAL_CMD_SEND) == 0)
 		err = serve_write(infd, outfd, gotd_sock, repo_path,
 		    chattygot);
 	else
